# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import Optional

import numpy as np

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.validation import ShapeT
from boulderopal._typing import Annotated
from boulderopal._validation import (
    ArrayT,
    Checker,
    ScalarT,
    pipe,
    validated,
)

from .node_data import (
    FilterFunction,
    Pwc,
    Tensor,
)


def _check_filter_function_input_shapes(
    control_hamiltonian: Pwc,
    noise_operator: Pwc,
    projection_operator: Optional[np.ndarray],
) -> None:
    Checker.VALUE(
        noise_operator.value_shape == control_hamiltonian.value_shape,
        "The noise operator must have same dimension as the control Hamiltonian.",
        {
            "control_hamiltonian.value_shape": control_hamiltonian.value_shape,
            "noise_operator.value_shape": noise_operator.value_shape,
        },
    )

    if projection_operator is not None:
        Checker.VALUE(
            projection_operator.shape == control_hamiltonian.value_shape,
            "The projection operator must have the same dimension as the control Hamiltonian.",
            {
                "control_hamiltonian.value_shape": control_hamiltonian.value_shape,
                "projection_operator.shape": projection_operator.shape,
            },
        )


class FilterFunctionGraph:
    """
    Base class implementing filter function graph methods.
    """

    @validated
    def filter_function(
        self,
        control_hamiltonian: Annotated[Pwc, pipe(after=ShapeT.OPERATOR().no_batch())],
        noise_operator: Annotated[Pwc, pipe(after=ShapeT.OPERATOR().no_batch())],
        frequencies: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1))],
        sample_count: Optional[Annotated[int, pipe(ScalarT.INT().gt(1))]] = 100,
        projection_operator: Optional[
            Annotated[
                np.ndarray,
                pipe(ArrayT.NUMERIC().ndim(2).square().hermitian().idempotent()),
            ]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> FilterFunction:
        r"""
        Evaluate the filter function for a control Hamiltonian and a noise operator
        at the given frequency elements.

        Parameters
        ----------
        control_hamiltonian : Pwc
            The control Hamiltonian :math:`H_\mathrm{c}(t)`.
        noise_operator : Pwc
            The noise operator :math:`N(t)`.
        frequencies : list or tuple or np.ndarray
            The elements in the frequency domain at which to return
            the values of the filter function.
        sample_count : int or None, optional
            The number of points in time, :math:`M`, to sample the control-frame noise operator.
            These samples are used to calculate the approximate Fourier integral efficiently.
            If None the piecewise Fourier integral is calculated exactly.
            Defaults to 100.
        projection_operator : np.ndarray or None, optional
            The projection operator :math:`P`. Defaults to the identity matrix.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        ~graphs.FilterFunction
            The filter function.

        See Also
        --------
        Graph.frequency_domain_noise_operator :
            Control-frame noise operator in the frequency domain.

        Notes
        -----
        The filter function is defined as [1]_:

        .. math::
            F(f) = \frac{1}{\mathrm{Tr}(P)} \mathrm{Tr} \left(
            P \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f)
            \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f)^\dagger P \right),

        with the control-frame noise operator in the frequency domain

        .. math::
            \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f) = \int_0^\tau
            e^{-i 2\pi f t} \tilde N^\prime(t) \mathrm{d}t,

        where

        .. math::
            \tilde N^\prime(t) = \tilde N(t) - \frac{\mathrm{Tr}\left( P \tilde N(t)
            P \right)}{\mathrm{Tr}(P)} \mathbb{I}

        is the traceless control-frame noise operator in the time domain,

        .. math::
            \tilde N(t) = U_c^\dagger(t) N(t) U_c(t)

        is the control-frame noise operator in the time domain, and :math:`U_c(t)` is
        the time evolution induced by the control Hamiltonian.

        References
        ----------
        .. [1] `H. Ball, M. J. Biercuk, A. R. R. Carvalho, J. Chen, M. Hush,
                L. A. De Castro, L. Li, P. J. Liebermann, H. J. Slatyer, and C. Edmunds,
                Quantum Sci. Technol. 6, 044011 (2021).
                <https://doi.org/10.1088/2058-9565/abdca6>`_
        """

        _check_filter_function_input_shapes(
            control_hamiltonian,
            noise_operator,
            projection_operator,
        )

        operation = create_operation(self.filter_function, locals())
        return FilterFunction(operation, frequencies=frequencies, exact=sample_count is None)

    @validated
    def frequency_domain_noise_operator(
        self,
        control_hamiltonian: Annotated[Pwc, pipe(after=ShapeT.OPERATOR().no_batch())],
        noise_operator: Annotated[Pwc, pipe(after=ShapeT.OPERATOR().no_batch())],
        frequencies: Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1))],
        sample_count: Optional[Annotated[int, pipe(ScalarT.INT().gt(1))]] = 100,
        projection_operator: Optional[
            Annotated[
                np.ndarray,
                pipe(ArrayT.NUMERIC().ndim(2).square().hermitian().idempotent()),
            ]
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a control-frame noise operator in the frequency domain for a control Hamiltonian
        and a noise operator at the given frequencies.

        Parameters
        ----------
        control_hamiltonian : Pwc
            The control Hamiltonian :math:`H_\mathrm{c}(t)`.
        noise_operator : Pwc
            The noise operator :math:`N(t)`.
        frequencies : list or tuple or np.ndarray
            The elements in the frequency domain at which to return
            the values of the filter function.
        sample_count : int or None, optional
            The number of points in time, :math:`M`, to sample the control-frame noise operator.
            These samples are used to calculate the approximate Fourier integral efficiently.
            If None the piecewise Fourier integral is calculated exactly.
            Defaults to 100.
        projection_operator : np.ndarray or None, optional
            The projection operator :math:`P`. Defaults to the identity matrix.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The noise operator in the frequency domain.

        See Also
        --------
        Graph.filter_function :
            The filter function.

        Notes
        -----
        The control-frame noise operator in the frequency domain
        is defined as the Fourier transform of the operator in the time domain [1]_:

        .. math::
            \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f) = \int_0^\tau
            e^{-i 2\pi f t} \tilde N^\prime(t) \mathrm{d}t,

        where

        .. math::
            \tilde N^\prime(t) = \tilde N(t) - \frac{\mathrm{Tr}\left( P \tilde N(t)
            P \right)}{\mathrm{Tr}(P)} \mathbb{I}

        is the traceless control-frame noise operator in the time domain,

        .. math::
            \tilde N(t) = U_c^\dagger(t) N(t) U_c(t)

        is the control-frame noise operator in the time domain, and :math:`U_c(t)` is
        the time evolution induced by the control Hamiltonian. If `sample_count` is
        set, the Fourier integral is approximated as

        .. math::
            \mathcal{F} \left\{ \tilde N^\prime(t) \right\}(f) \approx \sum_{n=0}^{M-1}
            \frac{\tau}{M} e^{-i 2\pi f n \tau/M}
            \langle \tilde N^\prime (n\tau/M) \rangle,

        where :math:`\tau` is the duration of the control Hamiltonian.

        References
        ----------
        .. [1] `H. Ball, M. J. Biercuk, A. R. R. Carvalho, J. Chen, M. Hush,
                L. A. De Castro, L. Li, P. J. Liebermann, H. J. Slatyer, and C. Edmunds,
                Quantum Sci. Technol. 6, 044011 (2021).
                <https://doi.org/10.1088/2058-9565/abdca6>`_
        """

        count = len(frequencies)
        _check_filter_function_input_shapes(
            control_hamiltonian,
            noise_operator,
            projection_operator,
        )

        operation = create_operation(self.frequency_domain_noise_operator, locals())
        return Tensor(operation, shape=(count,) + control_hamiltonian.value_shape)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from enum import Enum
from functools import partial
from typing import (
    Any,
    Callable,
    Tuple,
    TypeVar,
)

import numpy as np
from scipy.sparse import spmatrix

from boulderopal._nodes.node_data import (
    Pwc,
    SparsePwc,
    Stf,
    Target,
    Tensor,
)
from boulderopal._typing import Self
from boulderopal._validation import (
    ArrayT,
    Checker,
    ScalarT,
    type_pipe,
)

_S = TypeVar("_S", Pwc, SparsePwc, Stf, Target, Tensor, np.ndarray, spmatrix)


def shapeable(value: _S, *, name: str) -> _S:
    """
    Perform preliminary validation for the input:
        - ensure a spmatrix or Tensor input is not empty.
        - do nothing if it's a Boulder Opal type that contains shape information.
        - try to convert the input to NumPy array.
    Note that this function doesn't know the argument type, which is left for Pydantic to check.
    """
    if isinstance(value, (Tensor, spmatrix)):
        Checker.VALUE(np.prod(value.shape) != 0, f"The {name} can't have empty value.")
        return value
    if isinstance(value, (Pwc, SparsePwc, Stf, Target)):
        return value
    return ArrayT.NUMERIC(name)(value)


BatchValueShapes = Tuple[Tuple[int, ...], Tuple[int, ...]]


def _get_matrix_shape(value: Any, name: str) -> BatchValueShapes:
    def _validate_shape(x: tuple[int, ...], strict: bool = False) -> tuple[int, ...]:
        if strict:
            Checker.VALUE(len(x) == 2, f"The value of {name} must be 2D.")
        else:
            Checker.VALUE(len(x) >= 2, f"The {name} must be at least 2D.")
        return x

    if isinstance(value, (np.ndarray, Tensor, spmatrix)):
        shape = _validate_shape(value.shape)
        return shape[:-2], shape[-2:]
    if isinstance(value, (Pwc, Stf)):
        value_shape = _validate_shape(value.value_shape, strict=True)
        return value.batch_shape, value_shape
    if isinstance(value, (SparsePwc, Target)):
        value_shape = _validate_shape(value.value_shape, strict=True)
        return (), value_shape
    raise TypeError(f"Unknown type for a matrix, got {type(value)}.")


def _get_operator_shape(value: Any, name: str) -> BatchValueShapes:
    def _validate_shape(x: tuple[int, ...], strict: bool = False) -> tuple[int, ...]:
        if strict:
            Checker.VALUE(len(x) == 2, f"The value of {name} must be 2D.")
        else:
            Checker.VALUE(len(x) >= 2, f"The {name} must be at least 2D.")
        Checker.VALUE(x[-1] == x[-2], f"The {name} must be a square in the last two dimensions.")
        return x

    if isinstance(value, (np.ndarray, Tensor, spmatrix)):
        shape = _validate_shape(value.shape)
        return shape[:-2], shape[-2:]
    if isinstance(value, (Pwc, Stf)):
        value_shape = _validate_shape(value.value_shape, strict=True)
        return value.batch_shape, value_shape
    if isinstance(value, (SparsePwc, Target)):
        value_shape = _validate_shape(value.value_shape, strict=True)
        return (), value_shape
    raise TypeError(f"Unknown type for an operator, got {type(value)}.")


def _get_vector_shape(value: Any, name: str) -> BatchValueShapes:
    def _validate_shape(shape_: tuple[int, ...], strict: bool = False) -> tuple[int, ...]:
        Checker.VALUE(len(shape_) >= 1, f"The {name} must be at least 1D.")
        if strict:
            Checker.VALUE(len(shape_) == 1, f"The value of {name} must be 1D.")
        return shape_

    if isinstance(value, (np.ndarray, Tensor, spmatrix)):
        shape = _validate_shape(value.shape)
        return shape[:-1], shape[-1:]
    if isinstance(value, (Pwc, Stf)):
        value_shape = _validate_shape(value.value_shape, strict=True)
        return value.batch_shape, value_shape
    if isinstance(value, SparsePwc):
        value_shape = _validate_shape(value.value_shape, strict=True)
        return (), value_shape
    raise TypeError(f"Unknown type for a vector, got {type(value)}.")


def _get_signal_shape(value: Any, name: str) -> BatchValueShapes:
    def _validate_shape(shape_: tuple[int, ...]) -> tuple[int, ...]:
        Checker.VALUE(shape_ == (), f"The {name} must have scalar value.")
        return shape_

    if isinstance(value, (Pwc, Stf)):
        return value.batch_shape, _validate_shape(value.value_shape)
    raise TypeError(f"Unknown type for a signal, got {type(value)}.")


class ShapeT(Enum):
    """
    Validator to check the shape of types that can represent matrix, operator, vector, or signal.
    """

    MATRIX = partial(_get_matrix_shape)
    OPERATOR = partial(_get_operator_shape)
    VECTOR = partial(_get_vector_shape)
    SIGNAL = partial(_get_signal_shape)

    def __call__(self) -> _ShapeT:
        return _ShapeT(self)


class _ShapeT:
    """
    A validator class to check shapes for ShapeT.
    """

    def __init__(self, type_: ShapeT) -> None:
        self._type = type_
        self._name = ""
        self._func_pipe: list[Callable] = []
        self._value_dim = 0

    def batch(self, count: int) -> Self:
        """
        len(batch_dimension) <= count.
        """

        def _check(x: tuple[int, ...], _: Any) -> None:
            Checker.VALUE(
                len(x) <= count,
                f"{self._name} can have at most {count} batch dimensions.",
            )

        self._func_pipe.append(_check)
        return self

    def no_batch(self) -> Self:
        """
        Value can't have batch.
        """

        def _check(x: tuple[int, ...], _: Any) -> None:
            Checker.VALUE(
                x == (),
                f"The {self._name} doesn't support batched input, "
                f"its whole shape must be {self._value_dim}D.",
            )

        self._func_pipe.append(_check)
        return self

    def __call__(self, value: _S, *, name: str) -> _S:
        self._name = name
        batch_shape, value_shape = self._type.value(value, name=self._name)
        self._value_dim = len(value_shape)
        for f in self._func_pipe:
            f(batch_shape, value_shape)
        return value


def bounded_by(value: np.ndarray, name: str, bound: float, bound_name: str) -> np.ndarray:
    """
    value[-1] <= bound.
    """
    Checker.VALUE(
        value[-1] <= bound or np.isclose(value[-1], bound, atol=0.0),
        f"The {name} must be smaller than or equal to {bound_name}.",
    )
    return value


def starts_with_zero(value: np.ndarray, *, name: str) -> np.ndarray:
    """
    1D array should start with 0.
    """
    Checker.VALUE(np.isclose(value[0], 0), f"The {name} must start with zero.")
    return value


_T = TypeVar("_T", np.ndarray, Tensor)


def no_scalar(value: _T, *, name: str) -> _T:
    """
    Do not allow scalar array.
    """
    Checker.TYPE(len(value.shape) > 0, f"The {name} must not be scalar.")
    return value


def strict_real_array(value: _T, *, name: str) -> _T:
    """
    Ensure NumPy array is real.
    """
    if isinstance(value, np.ndarray):
        Checker.TYPE(value.dtype.kind in "iuf", f"The {name} must be a real array.")
    return value


def _to_scalar_tensor(value: Tensor, *, name: str) -> Tensor:
    """
    Check if a Tensor value contains a single element and return a scalar Tensor.
    """

    if isinstance(value, Tensor):
        if value.shape == (1,):
            return value[0]
        if value.shape == ():
            return value
    raise TypeError(f"The {name} must be a scalar Tensor.")


scalar_or_shapeable = type_pipe(
    [ScalarT.NUMERIC(), shapeable],
    messenger=lambda x: (
        f"The {x} must be either a non-empty NumPy array or Tensor, or a numeric scalar."
    ),
)


def _qutip_obj(value: Any, *, name: str) -> np.ndarray:
    if hasattr(value, "full"):
        return value.full()
    raise TypeError(f"{name} is not a QuTiP object.")


number_or_shapeable = type_pipe(
    [_qutip_obj, ScalarT.NUMERIC(), shapeable],
    messenger=lambda name: f"The variable {name} must be "
    "a number, a NumPy array, a Pwc, an Stf, or a Tensor.",
)


def _real_messenger(name: str) -> str:
    return (
        f"The variable {name} must be "
        "a real number, a real NumPy array, a Pwc, an Stf, or a real Tensor."
    )


real_shapeable = type_pipe([_qutip_obj, ScalarT.REAL(), shapeable], _real_messenger)
real_array = type_pipe([strict_real_array], messenger=_real_messenger)


def _scalar_messenger(name: str) -> str:
    return f"The {name} must be a scalar or a scalar-like Tensor."


scalar = type_pipe([ScalarT.NUMERIC(), _to_scalar_tensor], _scalar_messenger)
positive_scalar = type_pipe([ScalarT.REAL().gt(0), _to_scalar_tensor], _scalar_messenger)
non_negative_scalar = type_pipe([ScalarT.REAL().ge(0), _to_scalar_tensor], _scalar_messenger)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import (
    Any,
    Callable,
    List,
    Optional,
    Union,
)

import numpy as np
from pydantic import SkipValidation

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.node_data import (
    Pwc,
    Stf,
    Tensor,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    ArrayT,
    Checker,
    ScalarT,
    pipe,
    sequence_like,
    validated,
)


def _validate_initial_values(
    values: Any,
    count: int,
    lower: float,
    upper: float,
) -> Optional[np.ndarray | list[np.ndarray]]:
    Checker.VALUE(
        upper > lower,
        "The lower bound must be less than the upper bound.",
        {"lower_bound": lower, "upper_bound": upper},
    )
    if values is None:
        return None

    def _validate_value(value: Any, *, name: str) -> np.ndarray:
        return ArrayT.REAL(name).ndim(1).shape((count,)).ge(lower).le(upper)(value)

    try:
        return sequence_like(_validate_value)(values, name="initial_values")
    except TypeError as exc:
        raise TypeError(
            "The initial values must be a NumPy array or a list of NumPy arrays.",
        ) from exc


def _check_inputs_real_fourier_signal(
    lower: float,
    upper: float,
    fixed_frequencies: Optional[np.ndarray],
    optimizable_frequency_count: Optional[int],
    randomized_frequency_count: Optional[int],
) -> None:
    """
    Check if the inputs of real_fourier_pwc/stf_signal function are valid.
    """
    Checker.VALUE(
        upper > lower,
        "The initial coefficient lower bound must be less than the upper bound.",
        {
            "initial_coefficient_lower_bound": lower,
            "initial_coefficient_upper_bound": upper,
        },
    )
    Checker.VALUE(
        (fixed_frequencies is not None)
        + (optimizable_frequency_count is not None)
        + (randomized_frequency_count is not None)
        == 1,
        "Exactly one of `fixed_frequencies`, `optimizable_frequency_count`, and "
        "`randomized_frequency_count` must be provided.",
        {
            "fixed_frequencies": fixed_frequencies,
            "optimizable_frequency_count": optimizable_frequency_count,
            "randomized_frequency_count": randomized_frequency_count,
        },
    )


class OptimizationGraph:
    """
    Base class implementing optimization graph methods.
    """

    @validated
    def optimization_variable(
        self,
        count: Annotated[int, pipe(ScalarT.INT().gt(0))],
        lower_bound: Annotated[float, pipe(ScalarT.REAL())],
        upper_bound: Annotated[float, pipe(ScalarT.REAL())],
        is_lower_unbounded: bool = False,
        is_upper_unbounded: bool = False,
        initial_values: Annotated[
            Optional[Union[np.ndarray, List[np.ndarray]]],
            SkipValidation,
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a 1D Tensor of optimization variables, which can be bounded,
        semi-bounded, or unbounded.

        Use this function to create a sequence of variables that can be tuned by
        the optimizer (within specified bounds) in order to minimize the cost
        function.

        Parameters
        ----------
        count : int
            The number :math:`N` of individual real-valued variables to create.
        lower_bound : float
            The lower bound :math:`v_\mathrm{min}` for generating an initial value for the
            variables. This will also be used as lower bound if the variables are lower bounded.
            The same lower bound applies to all `count` individual variables.
        upper_bound : float
            The upper bound :math:`v_\mathrm{max}` for generating an initial value for the
            variables. This will also be used as upper bound if the variables are upper bounded.
            The same upper bound applies to all `count` individual variables.
        is_lower_unbounded : bool, optional
            Defaults to False. Set this flag to True to define a semi-bounded variable with
            lower bound :math:`-\infty`; in this case, the `lower_bound` parameter is used only for
            generating an initial value.
        is_upper_unbounded : bool, optional
            Defaults to False. Set this flag to True to define a semi-bounded variable with
            upper bound :math:`+\infty`; in this case, the `upper_bound` parameter is used only for
            generating an initial value.
        initial_values : np.ndarray or List[np.ndarray] or None, optional
            Initial values for the optimization variable. You can either provide a single initial
            value, or a list of them. Note that all optimization variables in a graph with
            non-default initial values must have the same length. That is, you must set them either
            as a single array or a list of arrays of the same length. Defaults to None, meaning
            the optimizer initializes the variables with random values.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The sequence :math:`\{v_n\}` of :math:`N` optimization variables. If both
            `is_lower_unbounded` and `is_upper_unbounded` are False, these variables are
            bounded such that :math:`v_\mathrm{min}\leq v_n\leq v_\mathrm{max}`. If one of the
            flags is True (for example `is_lower_unbounded=True`), these variables are
            semi-bounded (for example :math:`-\infty \leq v_n \leq v_\mathrm{max}`).
            If both of them are True, then these variables are unbounded and satisfy that
            :math:`-\infty \leq v_n \leq +\infty`.

        See Also
        --------
        Graph.anchored_difference_bounded_variables :
            Create anchored optimization variables with a difference bound.
        Graph.complex_optimizable_pwc_signal :
            Create a complex optimizable `Pwc` signal.
        Graph.optimizable_scalar : Create an optimization scalar.
        Graph.real_optimizable_pwc_signal :
            Create a real optimizable `Pwc` signal.
        boulderopal.run_optimization :
            Function to find the minimum of a generic function.

        Examples
        --------
        Perform a simple optimization task.

        >>> variables = graph.optimization_variable(
        ...     2, lower_bound=0, upper_bound=1, name="variables"
        ... )
        >>> x = variables[0]
        >>> y = variables[1]
        >>> cost = (x - 0.1) ** 2 + graph.sin(y) ** 2
        >>> cost.name = "cost"
        >>> result = bo.run_optimization(
        ...     graph=graph, cost_node_name="cost", output_node_names="variables"
        ... )
        >>> result["cost"]
        0.0
        >>> result["output"]["variables"]["value"]
        array([0.1, 0.])

        See examples about optimal control of quantum systems in the
        `How to optimize controls in arbitrary quantum systems using graphs
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-
        in-arbitrary-quantum-systems-using-graphs>`_ user guide.
        """

        initial_values = _validate_initial_values(initial_values, count, lower_bound, upper_bound)
        operation = create_operation(self.optimization_variable, locals())
        return Tensor(operation, shape=(count,))

    @validated
    def anchored_difference_bounded_variables(
        self,
        count: Annotated[int, pipe(ScalarT.INT().gt(0))],
        lower_bound: Annotated[float, pipe(ScalarT.REAL())],
        upper_bound: Annotated[float, pipe(ScalarT.REAL())],
        difference_bound: Annotated[float, pipe(ScalarT.REAL().ge(0))],
        initial_values: Annotated[
            Optional[Union[np.ndarray, List[np.ndarray]]],
            SkipValidation,
        ] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a sequence of optimizable variables with an anchored difference bound.

        Use this function to create a sequence of optimization variables that have
        a difference bound (each variable is constrained to be within a given
        distance of the adjacent variables) and are anchored to zero at the start
        and end (the initial and final variables are within a given distance of
        zero).

        Parameters
        ----------
        count : int
            The number :math:`N` of individual real-valued variables to create.
        lower_bound : float
            The lower bound :math:`v_\mathrm{min}` on the variables.
            The same lower bound applies to all `count` individual variables.
        upper_bound : float
            The upper bound :math:`v_\mathrm{max}` on the variables.
            The same upper bound applies to all `count` individual variables.
        difference_bound : float
            The difference bound :math:`\delta` to enforce between adjacent
            variables.
        initial_values : np.ndarray or List[np.ndarray] or None, optional
            Initial values for optimization variable. You can either provide a single initial
            value, or a list of them. Note that all optimization variables in a graph with
            non-default initial values must have the same length. That is, you must set them either
            as a single array or a list of arrays of the same length. Defaults to None, meaning
            the optimizer initializes the variables with random values.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The sequence :math:`\{v_n\}` of :math:`N` anchored difference-bounded
            optimization variables, satisfying
            :math:`v_\mathrm{min}\leq v_n\leq v_\mathrm{max}`,
            :math:`|v_{n-1}-v_n|\leq\delta` for :math:`2\leq n\leq N`,
            :math:`|v_1|\leq\delta`, and :math:`|v_N|\leq\delta`.

        See Also
        --------
        Graph.optimization_variable : Create 1D Tensor of optimization variables.
        boulderopal.run_optimization :
            Function to find the minimum of generic deterministic functions.
        boulderopal.run_stochastic_optimization :
            Function to find the minimum of generic stochastic functions.

        Examples
        --------
        Create optimizable PWC signal with anchored difference bound.

        >>> values = graph.anchored_difference_bounded_variables(
        ...     count=10, lower_bound=-1, upper_bound=1, difference_bound=0.1
        ... )
        >>> graph.pwc_signal(values=values, duration=1)
        <Pwc: name="pwc_signal_#1", operation_name="pwc_signal", value_shape=(), batch_shape=()>

        See the "Band-limited pulses with bounded slew rates" example in the
        `How to add smoothing and band-limits to optimized controls
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-add-smoothing-and-band-limits-to-
        optimized-controls#example-band-limited-pulses-with-bounded-slew-rates>`_ user guide.
        """

        initial_values = _validate_initial_values(initial_values, count, lower_bound, upper_bound)
        operation = create_operation(self.anchored_difference_bounded_variables, locals())
        return Tensor(operation, shape=(count,))

    @validated
    def real_fourier_stf_signal(
        self,
        duration: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        initial_coefficient_lower_bound: Annotated[float, pipe(ScalarT.REAL())] = -1,
        initial_coefficient_upper_bound: Annotated[float, pipe(ScalarT.REAL())] = 1,
        fixed_frequencies: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).gt(0))]
        ] = None,
        optimizable_frequency_count: Optional[Annotated[int, pipe(ScalarT.INT().gt(0))]] = None,
        randomized_frequency_count: Optional[Annotated[int, pipe(ScalarT.INT().gt(0))]] = None,
    ) -> Stf:
        r"""
        Create a real sampleable signal constructed from Fourier components.

        Use this function to create a signal defined in terms of Fourier (sine/cosine)
        basis signals that can be optimized by varying their coefficients and, optionally,
        their frequencies.

        Parameters
        ----------
        duration : float
            The total duration :math:`\tau` of the signal.
        initial_coefficient_lower_bound : float, optional
            The lower bound :math:`c_\mathrm{min}` on the initial coefficient
            values. Defaults to -1.
        initial_coefficient_upper_bound : float, optional
            The upper bound :math:`c_\mathrm{max}` on the initial coefficient
            values. Defaults to 1.
        fixed_frequencies : np.ndarray or None, optional
            A 1D array containing the fixed non-zero frequencies :math:`\{f_m\}`
            to use for the Fourier basis. If provided, must be non-empty and specified in
            the inverse units of `duration` (for example if `duration` is in seconds, these
            values must be given in Hertz).
        optimizable_frequency_count : int or None, optional
            The number of non-zero frequencies :math:`M` to use, if the
            frequencies can be optimized. Defaults to 0.
        randomized_frequency_count : int or None, optional
            The number of non-zero frequencies :math:`M` to use, if the
            frequencies are to be randomized but fixed. Defaults to 0.

        Returns
        -------
        Stf
            The optimizable, real-valued, sampleable signal built from the
            appropriate Fourier components.

        Warnings
        --------
        You must provide exactly one of `fixed_frequencies`, `optimizable_frequency_count`,
        or `randomized_frequency_count`.

        See Also
        --------
        Graph.real_fourier_pwc_signal : Corresponding operation for `Pwc`.

        Notes
        -----
        This function sets the basis signal frequencies :math:`\{f_m\}`
        depending on the chosen mode:

        * For fixed frequencies, you provide the frequencies directly.
        * For optimizable frequencies, you provide the number of frequencies
          :math:`M`, and this function creates :math:`M` unbounded optimizable
          variables :math:`\{f_m\}`, with initial values in the ranges
          :math:`\{[(m-1)/\tau, m/\tau]\}`.
        * For randomized frequencies, you provide the number of frequencies
          :math:`M`, and this function creates :math:`M` randomized constants
          :math:`\{f_m\}` in the ranges :math:`\{[(m-1)/\tau, m/\tau]\}`.

        After this function creates the :math:`M` frequencies :math:`\{f_m\}`, it
        produces the signal

        .. math::
            \alpha^\prime(t) = v_0 +
            \sum_{m=1}^M [ v_m \cos(2\pi t f_m) + w_m \sin(2\pi t f_m) ],

        where :math:`\{v_m,w_m\}` are (unbounded) optimizable variables, with
        initial values bounded by :math:`c_\mathrm{min}` and
        :math:`c_\mathrm{max}`. This function produces the final signal :math:`\alpha(t)`.

        You can use the signals created by this function for chopped random basis
        (CRAB) optimization [1]_.

        References
        ----------
        .. [1] `P. Doria, T. Calarco, and S. Montangero,
                Phys. Rev. Lett. 106, 190501 (2011).
                <https://doi.org/10.1103/PhysRevLett.106.190501>`_
        """

        _check_inputs_real_fourier_signal(
            initial_coefficient_lower_bound,
            initial_coefficient_upper_bound,
            fixed_frequencies,
            optimizable_frequency_count,
            randomized_frequency_count,
        )
        operation = create_operation(self.real_fourier_stf_signal, locals())
        return Stf(operation, value_shape=(), batch_shape=())

    @validated
    def real_fourier_pwc_signal(
        self,
        duration: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        segment_count: Annotated[int, pipe(ScalarT.INT().gt(0))],
        initial_coefficient_lower_bound: Annotated[float, pipe(ScalarT.REAL())] = -1,
        initial_coefficient_upper_bound: Annotated[float, pipe(ScalarT.REAL())] = 1,
        fixed_frequencies: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).gt(0))]
        ] = None,
        optimizable_frequency_count: Optional[Annotated[int, pipe(ScalarT.INT().gt(0))]] = None,
        randomized_frequency_count: Optional[Annotated[int, pipe(ScalarT.INT().gt(0))]] = None,
        *,
        name: Optional[str] = None,
    ) -> Pwc:
        r"""
        Create a piecewise-constant signal constructed from Fourier components.

        Use this function to create a signal defined in terms of Fourier
        (sine/cosine) basis signals that can be optimized by varying their
        coefficients and, optionally, their frequencies.

        Parameters
        ----------
        duration : float
            The total duration :math:`\tau` of the signal.
        segment_count : int
            The number of segments :math:`N` to use for the piecewise-constant
            approximation.
        initial_coefficient_lower_bound : float, optional
            The lower bound :math:`c_\mathrm{min}` on the initial coefficient
            values. Defaults to -1.
        initial_coefficient_upper_bound : float, optional
            The upper bound :math:`c_\mathrm{max}` on the initial coefficient
            values. Defaults to 1.
        fixed_frequencies : np.ndarray or None, optional
            A 1D array object containing the fixed non-zero frequencies :math:`\{f_m\}`
            to use for the Fourier basis. If provided, must be non-empty and specified in
            the inverse units of `duration` (for example if `duration` is in seconds, these
            values must be given in Hertz).
        optimizable_frequency_count : int or None, optional
            The number of non-zero frequencies :math:`M` to use, if the
            frequencies can be optimized. Defaults to 0.
        randomized_frequency_count : int or None, optional
            The number of non-zero frequencies :math:`M` to use, if the
            frequencies are to be randomized but fixed. Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Pwc
            The optimizable, real-valued, piecewise-constant signal built from the
            appropriate Fourier components.

        Warnings
        --------
        You must provide exactly one of `fixed_frequencies`, `optimizable_frequency_count`,
        or `randomized_frequency_count`.

        See Also
        --------
        Graph.real_fourier_stf_signal : Corresponding operation for `Stf`.

        Notes
        -----
        This function sets the basis signal frequencies :math:`\{f_m\}`
        depending on the chosen mode:

        * For fixed frequencies, you provide the frequencies directly.
        * For optimizable frequencies, you provide the number of frequencies
          :math:`M`, and this function creates :math:`M` unbounded optimizable
          variables :math:`\{f_m\}`, with initial values in the ranges
          :math:`\{[(m-1)/\tau, m/\tau]\}`.
        * For randomized frequencies, you provide the number of frequencies
          :math:`M`, and this function creates :math:`M` randomized constants
          :math:`\{f_m\}` in the ranges :math:`\{[(m-1)/\tau, m/\tau]\}`.

        After this function creates the :math:`M` frequencies :math:`\{f_m\}`, it
        produces the signal

        .. math::
            \alpha^\prime(t) = v_0 +
            \sum_{m=1}^M [ v_m \cos(2\pi t f_m) + w_m \sin(2\pi t f_m) ],

        where :math:`\{v_m,w_m\}` are (unbounded) optimizable variables, with
        initial values bounded by :math:`c_\mathrm{min}` and
        :math:`c_\mathrm{max}`. This function produces the final
        piecewise-constant signal :math:`\alpha(t)` by sampling
        :math:`\alpha^\prime(t)` at :math:`N` equally spaced points along the
        duration :math:`\tau`, and using those sampled values as the constant
        segment values.

        You can use the signals created by this function for chopped random basis
        (CRAB) optimization [1]_.

        References
        ----------
        .. [1] `P. Doria, T. Calarco, and S. Montangero,
                Phys. Rev. Lett. 106, 190501 (2011).
                <https://doi.org/10.1103/PhysRevLett.106.190501>`_

        Examples
        --------
        See the "Fourier-basis optimization on a qutrit" example in the
        `How to optimize controls using arbitrary basis functions
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-using-arbitrary-
        basis-functions#example-fourier-basis-optimization-on-a-qutrit>`_ user guide.
        """

        _check_inputs_real_fourier_signal(
            initial_coefficient_lower_bound,
            initial_coefficient_upper_bound,
            fixed_frequencies,
            optimizable_frequency_count,
            randomized_frequency_count,
        )

        durations = duration / segment_count * np.ones(segment_count)
        operation = create_operation(self.real_fourier_pwc_signal, locals())
        return Pwc(operation, value_shape=(), durations=durations, batch_shape=())

    @validated
    def optimizable_scalar(
        self,
        lower_bound: Annotated[float, pipe(ScalarT.REAL())],
        upper_bound: Annotated[float, pipe(ScalarT.REAL())],
        is_lower_unbounded: bool = False,
        is_upper_unbounded: bool = False,
        initial_values: Annotated[Optional[Union[float, List[float]]], SkipValidation] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create an optimizable scalar Tensor, which can be bounded, semi-bounded, or unbounded.

        Use this function to create a single variable that can be tuned by
        the optimizer to minimize the cost function.

        Parameters
        ----------
        lower_bound : float
            The lower bound :math:`v_\mathrm{min}` for generating an initial value for the variable.
            This will also be used as lower bound if the variable is lower bounded.
        upper_bound : float
            The upper bound :math:`v_\mathrm{max}` for generating an initial value for the variable.
            This will also be used as upper bound if the variable is upper bounded.
        is_lower_unbounded : bool, optional
            Defaults to False. Set this flag to True to define a semi-bounded variable with
            lower bound :math:`-\infty`; in this case, the `lower_bound` parameter is used only for
            generating an initial value.
        is_upper_unbounded : bool, optional
            Defaults to False. Set this flag to True to define a semi-bounded variable with
            upper bound :math:`+\infty`; in this case, the `upper_bound` parameter is used only for
            generating an initial value.
        initial_values : float or List[float] or None, optional
            Initial values for the optimization variable. You can either provide a single initial
            value, or a list of them. Note that all optimization variables in a graph with
            non-default initial values must have the same length. That is, you must set them all
            either as a single value or a list of values of the same length. Defaults to None,
            meaning the optimizer initializes the variable with a random value.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The :math:`v` optimizable scalar. If both `is_lower_unbounded` and `is_upper_unbounded`
            are False, the variables is bounded such that
            :math:`v_\mathrm{min}\leq v \leq v_\mathrm{max}`. If one of the flags is True
            (for example `is_lower_unbounded=True`), the variable is semi-bounded
            (for example :math:`-\infty \leq v \leq v_\mathrm{max}`). If both of them are True,
            then the variable is unbounded and satisfies that :math:`-\infty \leq v \leq +\infty`.

        See Also
        --------
        Graph.optimization_variable : Create 1D Tensor of optimization variables.
        boulderopal.run_optimization :
            Function to find the minimum of a generic function.
        """

        initial_array: Optional[np.ndarray | list[np.ndarray]] = None
        if initial_values is not None:

            def _validator(name: str) -> Callable:
                return ScalarT.REAL(name).ge(lower_bound).le(upper_bound)

            if isinstance(initial_values, list):
                initial_array = [
                    np.array([_validator(f"initial_values[{idx}]")(value)])
                    for idx, value in enumerate(initial_values)
                ]
            else:
                initial_array = np.array([_validator("initial_values")(initial_values)])

        scalar = self.optimization_variable(
            count=1,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            is_lower_unbounded=is_lower_unbounded,
            is_upper_unbounded=is_upper_unbounded,
            initial_values=initial_array,
        )[0]

        if name is not None:
            scalar.name = name

        return scalar



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import (
    Any,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
from scipy.sparse import (
    coo_matrix,
    issparse,
    spmatrix,
)

from boulderopal._nodes.base import create_operation
from boulderopal._nodes.validation import (
    ShapeT,
    bounded_by,
    shapeable,
)
from boulderopal._typing import Annotated
from boulderopal._validation import (
    ArrayT,
    Checker,
    ScalarT,
    pipe,
    validated,
)

from .node_data import (
    Pwc,
    SparsePwc,
    Tensor,
)

_LindbladTerms = Sequence[Tuple[float, Union[coo_matrix, Tensor, np.ndarray]]]


def _normalize_lindblad_terms(terms: Any, *, name: str) -> _LindbladTerms:
    """
    Validate types of lindblad_terms and normalize the type for each term.
    """
    Checker.TYPE(isinstance(terms, list) and len(terms) >= 0, f"The {name} must not be empty.")
    normalized = []
    for index, term in enumerate(terms):
        Checker.VALUE(
            isinstance(term, tuple) and len(term) == 2,
            f"The {name} must be a list of tuples with a decay rate and an operator each.",
            {f"Item {index} of {name}": term},
        )
        rate, op = term
        rate = ScalarT.REAL(f"decay rate in item {index} of {name}").gt(0)(rate)

        op_name = f"operator in item {index} of {name}"
        if isinstance(op, spmatrix):
            op = op.tocoo()
        elif not isinstance(op, Tensor):
            try:
                op = ArrayT.NUMERIC(op_name)(op)
            except TypeError as e:
                raise TypeError(
                    f"The {op_name} must be a NumPy array, or Tensor, or spmatrix.",
                ) from e
        op = ShapeT.OPERATOR()(op, name=op_name)

        normalized.append((rate, op))

    return normalized


def _check_hamiltonian_shape(hamiltonian: Pwc | SparsePwc, dimension: int, name: str) -> None:
    """
    Check if Hamiltonian has a compatible `dimension`.
    """
    Checker.VALUE(
        hamiltonian.value_shape == (dimension, dimension),
        f"The dimension of the Hamiltonian must be compatible with the dimension of {name}.",
        {
            "hamiltonian dimension": hamiltonian.value_shape,
            f"dimension of {name}": dimension,
        },
    )


def _check_lindblad_shape(lindblad_terms: _LindbladTerms, dimension: int, name: str) -> None:
    """
    Check if each Lindblad operator has the same `dimension`.
    """
    for index, (_, op) in enumerate(lindblad_terms):
        Checker.VALUE(
            op.shape[0] == dimension,
            f"The dimension of Lindblad operator of item {index} must be compatible "
            f"with the dimension of {name}.",
            {"Lindblad operator shape": op.shape, f"dimension of {name}": dimension},
        )


class OqsGraph:
    """
    Base class implementing OQS graph methods.
    """

    @validated
    def density_matrix_evolution_pwc(
        self,
        initial_density_matrix: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.OPERATOR().batch(1)),
        ],
        hamiltonian: Annotated[
            Union[Pwc, SparsePwc],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
        lindblad_terms: Annotated[
            Sequence[Tuple[float, Union[np.ndarray, spmatrix, Tensor]]],
            pipe(_normalize_lindblad_terms),
        ],
        sample_times: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).ge(0).ascend())]
        ] = None,
        error_tolerance: Optional[Annotated[float, pipe(ScalarT.REAL().le(1e-2))]] = 1e-6,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the state evolution of an open system described by the GKS–Lindblad master
        equation.

        The controls that you provide to this function have to be in piecewise-constant
        format. If your controls are smooth sampleable tensor-valued functions (STFs), you
        have to discretize them with `discretize_stf` before passing them to this function.
        You may need to increase the number of segments that you choose for the
        discretization depending on the sizes of oscillations in the smooth controls.

        By default, this function computes an approximate piecewise-constant solution for
        the consideration of efficiency, with the accuracy controlled by the
        parameter `error_tolerance`. If your system is small, you can set the
        `error_tolerance` to None to obtain an exact solution.

        Note that when using the exact method, both `hamiltonian` and `lindblad_terms`
        are converted to the dense representation, regardless of their original formats.
        This means that the computation can be slow and memory intensive when applied to
        large systems.

        When using the approximate method, the sparse representation is used internally if
        `hamiltonian` is a `SparsePwc`, otherwise the dense representation is used.

        Parameters
        ----------
        initial_density_matrix : np.ndarray or Tensor
            A 2D array of the shape ``(D, D)`` representing the initial density matrix of
            the system, :math:`\rho_{\mathrm s}`. You can also pass a batch of density matrices
            and the input data shape must be ``(B, D, D)`` where ``B`` is the batch dimension.
        hamiltonian : Pwc or SparsePwc
            A piecewise-constant function representing the effective system Hamiltonian,
            :math:`H_{\mathrm s}(t)`, for the entire evolution duration.
        lindblad_terms : list[tuple[float, np.ndarray or Tensor or scipy.sparse.spmatrix]]
            A list of pairs, :math:`(\gamma_j, L_j)`, representing the positive decay rate
            :math:`\gamma_j` and the Lindblad operator :math:`L_j` for each coupling
            channel :math:`j`. You must provide at least one Lindblad term.
        sample_times : np.ndarray or None, optional
            A 1D array like object of length :math:`T` specifying the times :math:`\{t_i\}` at
            which this function calculates system states. Must be ordered and contain at least
            one element. Note that increasing the density of sample times does not affect the
            computation precision of this function. If omitted only the evolved density matrix
            at the final time of the system Hamiltonian is returned.
        error_tolerance : float or None, optional
            Defaults to 1e-6. This option enables an approximate method to solve the master
            equation, meaning the 2-norm of the difference between the propagated state and the
            exact solution at the final time (and at each sample time if passed) is within the error
            tolerance. Note that, if set, this value must be smaller than 1e-2 (inclusive).
            However, setting it to a too small value (for example below 1e-12) might result in
            slower computation, but would not further improve the precision, since the dominating
            error in that case is due to floating point error. You can also explicitly set this
            option to None to find the exact piecewise-constant solution. Note that using the
            exact solution can be very computationally demanding in calculations involving a large
            Hilbert space or a large number of segments.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The time-evolved density matrix, with shape ``(D, D)`` or ``(T, D, D)``,
            depending on whether you provided sample times.
            If you provide a batch of initial states, the shape is
            ``(B, T, D, D)`` or ``(B, D, D)``.

        See Also
        --------
        Graph.jump_trajectory_evolution_pwc :
            Trajectory-based state evolution of an open quantum system.
        Graph.steady_state : Compute the steady state of open quantum system.

        Notes
        -----
        Under the Markovian approximation, the dynamics of an open quantum system can be
        described by the GKS–Lindblad master equation [1]_ [2]_

        .. math::
            \frac{{\mathrm d}\rho_{\mathrm s}(t)}{{\mathrm d}t} =
            -i [H_{\mathrm s}(t), \rho_{\mathrm s}(t)]
            + \sum_j \gamma_j {\mathcal D}[L_j] \rho_{\mathrm s}(t) ,

        where :math:`{\mathcal D}` is a superoperator describing the decoherent process in the
        system evolution and defined as

        .. math::
            {\mathcal D}[X]\rho := X \rho X^\dagger
                - \frac{1}{2}\left( X^\dagger X \rho + \rho X^\dagger X \right)

        for any system operator :math:`X`.

        This function uses sparse matrix multiplication when the Hamiltonian is passed as a
        `SparsePwc` and the Lindblad operators as sparse matrices. This leads to more efficient
        calculations when they involve large operators that are relatively sparse (contain mostly
        zeros). In this case, the initial density matrix is still a densely represented array
        or tensor.

        References
        ----------
        .. [1] `V. Gorini, A. Kossakowski, and E. C. G. Sudarshan,
                J. Math. Phys. 17, 821 (1976).
                <https://doi.org/10.1063/1.522979>`_
        .. [2] `G. Lindblad,
                Commun. Math. Phys. 48, 119 (1976).
                <https://doi.org/10.1007/BF01608499>`_

        Examples
        --------
        Simulate a trivial decay process for a single qubit described by the following
        master equation
        :math:`\dot{\rho} = -i[\sigma_z / 2, \, \rho] + \mathcal{D}[\sigma_-]\rho`.

        >>> duration = 20
        >>> initial_density_matrix = np.array([[0, 0], [0, 1]])
        >>> hamiltonian = graph.constant_pwc_operator(
        ...     duration=duration, operator=graph.pauli_matrix("Z") / 2
        ... )
        >>> lindblad_terms = [(1, graph.pauli_matrix("M"))]
        >>> graph.density_matrix_evolution_pwc(
        ...     initial_density_matrix, hamiltonian, lindblad_terms, name="decay"
        ... )
        <Tensor: name="decay", operation_name="density_matrix_evolution_pwc", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="decay")
        >>> result["output"]["decay"]["value"]
        array([[9.99999998e-01+0.j, 0.00000000e+00+0.j],
               [0.00000000e+00+0.j, 2.06115362e-09+0.j]])

        See more examples in the `How to simulate open system dynamics
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-open-system-dynamics>`_
        and `How to simulate large open system dynamics
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-
        large-open-system-dynamics>`_ user guides.
        """

        _shape = initial_density_matrix.shape
        _check_hamiltonian_shape(hamiltonian, _shape[-1], "initial_density_matrix")
        _check_lindblad_shape(lindblad_terms, _shape[-1], "initial_density_matrix")

        if sample_times is None:
            shape = _shape
        else:
            duration = np.sum(hamiltonian.durations)
            sample_times = bounded_by(
                sample_times,
                "sample_times",
                duration,
                "the duration of Hamiltonian",
            )
            shape = _shape[:-2] + (len(sample_times),) + _shape[-2:]
        operation = create_operation(self.density_matrix_evolution_pwc, locals())
        return Tensor(operation, shape=shape)

    @validated
    def jump_trajectory_evolution_pwc(
        self,
        initial_state_vector: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.VECTOR().batch(1)),
        ],
        hamiltonian: Annotated[
            Union[Pwc, SparsePwc],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
        lindblad_terms: Annotated[
            Sequence[Tuple[float, Union[np.ndarray, spmatrix, Tensor]]],
            pipe(_normalize_lindblad_terms),
        ],
        max_time_step: Annotated[float, pipe(ScalarT.REAL().gt(0))],
        trajectory_count: Annotated[int, pipe(ScalarT.INT().gt(0))],
        sample_times: Optional[
            Annotated[np.ndarray, pipe(ArrayT.REAL().ndim(1).ge(0).ascend())]
        ] = None,
        seed: Optional[Annotated[int, pipe(ScalarT.INT().ge(0))]] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the state evolution of an open system described by the GKS–Lindblad master
        equation using a jump-based trajectory method.

        This function calculates multiple pure-state trajectories starting from an initial pure
        state and returns the average density matrix over all trajectories.

        Note that regardless of their original formats, both `hamiltonian` and `lindblad_terms`
        are internally converted to a dense representation, so there is no computational advantage
        in using a sparse representation with this method.

        Parameters
        ----------
        initial_state_vector : np.ndarray or Tensor
            The initial state vector :math:`|\psi\rangle` as a ``(D,)`` array or Tensor,
            or batch of initial state vectors as a ``(B, D)`` array or Tensor.
        hamiltonian : Pwc or SparsePwc
            A piecewise-constant function representing the system Hamiltonian,
            :math:`H_{\mathrm s}(t)`, for the entire evolution duration,
            with Hilbert space dimension D.
        lindblad_terms : list[tuple[float, np.ndarray or Tensor or scipy.sparse.spmatrix]]
            A list of pairs, :math:`(\gamma_j, L_j)`, representing the positive decay rate
            :math:`\gamma_j` and the Lindblad operator :math:`L_j` for each coupling
            channel :math:`j`. You must provide at least one Lindblad term.
        max_time_step : float
            The maximum time step to use in the integration.
            Each PWC segment will be subdivided into steps that are, at most, this value.
            A smaller value for the maximum time step will more accurately sample the jump
            processes, but also lead to a slower computation.
        trajectory_count : int
            The number of quantum trajectories to run.
        sample_times : np.ndarray or None, optional
            A 1D array like object of length :math:`T` specifying the times :math:`\{t_i\}` at
            which this function calculates system states. Must be ordered and contain at least
            one element. If omitted only the evolved density matrix at the final time of the
            system Hamiltonian is returned. Note that increasing the density of sample times
            does not affect the computation precision of this function, but might slow down
            the calculation.
        seed : int or None, optional
            A seed for the random number generator.
            Defaults to None, in which case a random value for the seed is used.
        name : str, optional
            The name of the node.

        Returns
        -------
        Tensor
            The time-evolved density matrix, with shape ``(D, D)`` or ``(T, D, D)``,
            depending on whether you provided sample times.
            If you provide a batch of initial states, the shape is ``(B, T, D, D)``
            or ``(B, D, D)``.

        See Also
        --------
        Graph.density_matrix_evolution_pwc : State evolution of an open quantum system.
        Graph.steady_state : Compute the steady state of open quantum system.

        Notes
        -----
        Under the Markovian approximation, the dynamics of an open quantum system can be described
        by the GKS–Lindblad master equation [1]_ [2]_

        .. math::
            \frac{{\mathrm d}\rho_{\mathrm s}(t)}{{\mathrm d}t} =
            -i [H_{\mathrm s}(t), \rho_{\mathrm s}(t)]
            + \sum_j \gamma_j {\mathcal D}[L_j] \rho_{\mathrm s}(t) ,

        where :math:`{\mathcal D}` is a superoperator describing the decoherent process in the
        system evolution and defined as

        .. math::
            {\mathcal D}[X]\rho := X \rho X^\dagger
                - \frac{1}{2}\left( X^\dagger X \rho + \rho X^\dagger X \right)

        for any system operator :math:`X`.

        This function solves the GKS–Lindblad master equation with an initial pure state
        :math:`\rho_{\mathrm s}(0) = |\psi\rangle\langle\psi|`, by calculating
        multiple quantum trajectories performing quantum jumps,
        :math:`|\tilde\psi_k(t)\rangle`, and averaging the result:

        .. math::
            \rho(t) = \frac{1}{M} \sum_{k=1}^M |\tilde\psi_k(t) \rangle\langle \tilde\psi_k(t) | .

        References
        ----------
        .. [1] `V. Gorini, A. Kossakowski, and E. C. G. Sudarshan,
                J. Math. Phys. 17, 821 (1976).
                <https://doi.org/10.1063/1.522979>`_
        .. [2] `G. Lindblad,
                Commun. Math. Phys. 48, 119 (1976).
                <https://doi.org/10.1007/BF01608499>`_
        """

        dim = initial_state_vector.shape[-1]
        _check_hamiltonian_shape(hamiltonian, dim, "initial_density_matrix")
        _check_lindblad_shape(lindblad_terms, dim, "initial_density_matrix")

        shape: tuple[int, ...] = (dim, dim)
        if sample_times is not None:
            duration = np.sum(hamiltonian.durations)
            sample_times = bounded_by(
                sample_times,
                "sample_times",
                duration,
                "the duration of Hamiltonian",
            )
            shape = (len(sample_times),) + shape
        if len(initial_state_vector.shape) != 1:
            shape = (initial_state_vector.shape[0],) + shape

        operation = create_operation(self.jump_trajectory_evolution_pwc, locals())
        return Tensor(operation, shape=shape)

    @validated
    def steady_state(
        self,
        hamiltonian: Annotated[
            Union[Tensor, np.ndarray, spmatrix],
            pipe(shapeable, after=ShapeT.OPERATOR().no_batch()),
        ],
        lindblad_terms: Annotated[
            Sequence[Tuple[float, Union[np.ndarray, spmatrix, Tensor]]],
            pipe(_normalize_lindblad_terms),
        ],
        method: Literal["QR", "EIGEN_DENSE", "EIGEN_SPARSE"] = "QR",
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Find the steady state of a time-independent open quantum system.

        The Hamiltonian and Lindblad operators that you provide have to give
        rise to a unique steady state.

        Parameters
        ----------
        hamiltonian : Tensor or spmatrix
            A 2D array of shape ``(D, D)`` representing the time-independent
            Hamiltonian of the system, :math:`H_{\mathrm s}`.
        lindblad_terms : list[tuple[float, np.ndarray or Tensor or scipy.sparse.spmatrix]]
            A list of pairs, :math:`(\gamma_j, L_j)`, representing the positive decay rate
            :math:`\gamma_j` and the Lindblad operator :math:`L_j` for each coupling
            channel :math:`j`. You must provide at least one Lindblad term.
        method : str, optional
            The method used to find the steady state.
            Must be one of "QR", "EIGEN_SPARSE", or "EIGEN_DENSE".
            The "QR" method obtains the steady state through a QR decomposition and is suitable
            for small quantum systems with dense representation.
            The "EIGEN_SPARSE" and "EIGEN_DENSE" methods find the steady state as the eigenvector
            whose eigenvalue is closest to zero, using either a sparse or a dense representation
            of the generator.
            Defaults to "QR".
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The density matrix representing the steady state of the system.

        Warnings
        --------
        This function currently does not support calculating the gradient with respect to its
        inputs. Therefore, it cannot be used in a graph for a `run_optimization` or
        `run_stochastic_optimization` call, which will raise a `RuntimeError`.
        Please use gradient-free optimization if you want to perform an optimization task with this
        function. You can learn more about it in the
        `How to optimize controls using gradient-free optimization
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-controls-using-
        gradient-free-optimization>`_ user guide.

        See Also
        --------
        Graph.density_matrix_evolution_pwc : State evolution of an open quantum system.
        Graph.jump_trajectory_evolution_pwc :
            Trajectory-based state evolution of an open quantum system.

        Notes
        -----
        Under the Markovian approximation, the dynamics of an open quantum system can be described
        by the GKS–Lindblad master equation [1]_ [2]_,

        .. math::
            \frac{{\mathrm d}\rho_{\mathrm s}(t)}{{\mathrm d}t} =
            {\mathcal L} (\rho_{\mathrm s}(t)) ,

        where the Lindblad superoperator :math:`{\mathcal L}` is defined as

        .. math::
            {\mathcal L} (\rho_{\mathrm s}(t)) = -i [H_{\mathrm s}(t), \rho_{\mathrm s}(t)]
            + \sum_j \gamma_j {\mathcal D}[L_j] \rho_{\mathrm s}(t) ,

        where :math:`{\mathcal D}` is a superoperator describing the decoherent process in the
        system evolution and defined as

        .. math::
            {\mathcal D}[X]\rho := X \rho X^\dagger
                - \frac{1}{2}\left( X^\dagger X \rho + \rho X^\dagger X \right)

        for any system operator :math:`X`.

        This function computes the steady state of :math:`{\mathcal L}` by solving

        .. math:: \frac{{\mathrm d}\rho_{\mathrm s}(t)}{{\mathrm d}t} = 0 .

        The function assumes that :math:`H_{\mathrm s}` is time independent
        and that the dynamics generated by :math:`{\mathcal L}`
        give rise to a unique steady state. That is, the generated quantum dynamical map
        has to be ergodic [3]_.

        References
        ----------
        .. [1] `V. Gorini, A. Kossakowski, and E. C. G. Sudarshan,
                J. Math. Phys. 17, 821 (1976).
                <https://doi.org/10.1063/1.522979>`_
        .. [2] `G. Lindblad,
                Commun. Math. Phys. 48, 119 (1976).
                <https://doi.org/10.1007/BF01608499>`_
        .. [3] `D. Burgarth, G. Chiribella, V. Giovannetti, P. Perinotti, and K. Yuasa,
                New J. Phys. 15 073045 (2013).
                <https://doi.org/10.1088/1367-2630/15/7/073045>`_

        Examples
        --------
        Compute the steady state of the single qubit open system dynamics according to the
        Hamiltonian :math:`H=\omega\sigma_z` and the single Lindblad operator :math:`L=\sigma_-`.

        >>> omega = 0.8
        >>> gamma = 0.5
        >>> hamiltonian = omega * graph.pauli_matrix("Z")
        >>> lindblad_terms = [(gamma, graph.pauli_matrix("M"))]
        >>> graph.steady_state(hamiltonian, lindblad_terms, name="steady_state")
        <Tensor: name="steady_state", operation_name="steady_state", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="steady_state")
        >>> result["output"]["steady_state"]["value"]
        array([[1.+0.j 0.-0.j]
               [0.-0.j 0.-0.j]])
        """

        is_hermitian = True
        if issparse(hamiltonian):
            is_hermitian = np.allclose((hamiltonian - hamiltonian.getH()).data, 0)  # type: ignore
        elif isinstance(hamiltonian, np.ndarray):
            is_hermitian = np.allclose(hamiltonian, hamiltonian.T.conj())
        Checker.VALUE(is_hermitian, "The Hamiltonian must be Hermitian.")
        _check_lindblad_shape(lindblad_terms, hamiltonian.shape[-1], "hamiltonian")
        operation = create_operation(self.steady_state, locals())
        return Tensor(operation, shape=hamiltonian.shape)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from functools import (
    partial,
    wraps,
)
from itertools import chain
from typing import (
    Any,
    Callable,
    Iterator,
    List,
    Optional,
    Protocol,
    TypeVar,
    Union,
)

from pydantic import (
    ConfigDict,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapValidator,
    validate_call,
)

from boulderopal._typing import ParamSpec
from boulderopal._validation.exceptions import Checker

T = TypeVar("T")


class _ValidatorT(Protocol):
    """
    All validator function should satisfy this protocol.
    """

    def __call__(self, value: Any, *, name: str) -> Any: ...


_ValidatorPipe = Optional[Union[List[_ValidatorT], _ValidatorT]]


def pipe(before: _ValidatorPipe = None, *, after: _ValidatorPipe = None) -> WrapValidator:
    """
    This function is to mimic the behavior of the `Before/AfterValidator` with the `WrapValidator`
    such that one doesn't need to pass the name of the variable to be checked, and provide
    a simpler interface to define execution order.

    `before` and `after` are either a single validator or a list of them.
    All validators must be a callable with a signature like `func(value, *, name)`.

    The execution order is to run validators in the before list in the order as they are
    inserted, then invoke the Pydantic handler, and finally run all validators in the after list
    in the order they are inserted.
    """

    def inner_validator(value: T, handler: ValidatorFunctionWrapHandler, info: ValidationInfo) -> T:
        def _normalize(validator: _ValidatorPipe) -> Iterator[Callable[[T], T]]:
            assert isinstance(info.field_name, str)
            if validator is None:
                return iter(())
            if callable(validator):
                return iter([partial(validator, name=info.field_name)])
            return (partial(func, name=info.field_name) for func in validator)

        for _func in chain(_normalize(before), [handler], _normalize(after)):  # type: ignore
            value = _func(value)  # type: ignore
        return value

    return WrapValidator(inner_validator)


P = ParamSpec("P")
R = TypeVar("R")


def validated(func: Callable[P, R]) -> Callable[P, R]:
    """
    Decorator to mark a callable as validated by Pydantic (V2) and in-house validators.

    The function wraps `validate_call` and `func` so that we can use Pydantic
    to validate inputs for the callable and also maintain its signature
    and docstring.

    Note that we enable the `arbitrary_types_allowed` flag to support the
    customized/flexible types in Boulder Opal.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return validate_call(func, config=ConfigDict(arbitrary_types_allowed=True))(  # type: ignore
            *args,
            **kwargs,
        )

    return wrapper


def type_pipe(validators: list[_ValidatorT], messenger: Callable[[str], str]) -> _ValidatorT:
    """
    Run the validators in the `validators` list sequentially until the value is resolved.
    Otherwise, throw a TypeError with the message customized by the `messenger` callable.

    This could be useful when resolving union types like scalar | Tensor.
    Similarly, all validators must have a signature like `func(value, * , name)`.
    """

    def _inner(value: T, *, name: str) -> T:
        for _validator in validators:
            try:
                return _validator(value, name=name)
            except TypeError:
                pass
        raise TypeError(messenger(name))

    return _inner


def sequence_like(
    validator: _ValidatorT,
    normalizer: Optional[type] = None,
    min_length: int = 1,
) -> _ValidatorT:
    """
    Validate a single element or a list/tuple of elements with the same type.
    The element validator must have a signature like `func(value, * , name)`.

    This is useful in the case we want to perform validation for input type like `T | list[T]`,
    where we need to apply the same validator either to `T` or `list[T]`.

    Type like `T | list[T]` is often introduced for UX. If we want to normalize internally
    (i.e., to only handle `list[T]`), we can pass the optional `normalizer`. This parameter
    is expected to either be `list` or `tuple`.

    Note that similar to other utility Pydantic functions, this helper assumes that input
    can only be 1D sequence and doesn't support types other than list and tuple. Such
    type-checking is done by Pydantic.
    """

    def _inner(value: T, *, name: str) -> list[T] | tuple[T, ...] | T:
        if isinstance(value, (list, tuple)):
            Checker.VALUE(
                len(value) >= min_length,
                f"The {name} must have at least length {min_length}.",
            )
            return type(value)(
                validator(item, name=f"{name}[{idx}]") for idx, item in enumerate(value)
            )
        if normalizer is not None:
            return normalizer([validator(value, name=name)])
        return validator(value, name=name)

    return _inner



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

# Make private alias for all imports to clean up the autocomplete for the graph module.
import sys as _sys

import boulderopal._nodes.node_data as _nd
from boulderopal._nodes.ions import IonsGraph as _IonsGraph
from boulderopal._nodes.random import RandomGraph as _RandomGraph
from boulderopal._nodes.signals import Signals as _Signals
from boulderopal.graph._execute_graph import (
    ExecutionMode,
    execute_graph,
)
from boulderopal.graph._graph import Graph

_TYPE_REGISTRY = [
    _nd.ConvolutionKernel,
    _nd.FilterFunction,
    _nd.Pwc,
    _nd.SparsePwc,
    _nd.Stf,
    _nd.Target,
    _nd.Tensor,
]

# Binding the Graph related types to the current module, this is useful
# for building docs for the Graph class.
_module = _sys.modules[__name__]
for _type_cls in _TYPE_REGISTRY:
    setattr(_module, _type_cls.__name__, _type_cls)


# To build doc for namespace nodes. These are added as if they are part of the `Graph` module.
_module.random = _RandomGraph  # type: ignore[attr-defined]
_module.ions = _IonsGraph  # type: ignore[attr-defined]
_module.signals = _Signals  # type: ignore[attr-defined]


__all__ = ["ExecutionMode", "Graph", "execute_graph"]



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    Any,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

import numpy as np
from pydantic import Field

from boulderopal._typing import Annotated
from boulderopal._validation import (
    Checker,
    ScalarT,
    pipe,
    sequence_like,
    type_pipe,
    validated,
)

from .base import create_operation
from .node_data import Tensor
from .utils import (
    check_operation_axis,
    get_broadcasted_shape,
    get_keepdims_operation_shape,
)
from .validation import (
    ShapeT,
    no_scalar,
    scalar_or_shapeable,
    shapeable,
)


def _embed_messenger(name: str) -> str:
    return f"The {name} must be a list of tuples with an operator and an integer position each."


def _embed_tuple(value: Any, *, name: str) -> Tuple[Union[np.ndarray, Tensor], int]:
    Checker.TYPE(isinstance(value, (list, tuple)) and len(value) == 2, _embed_messenger(name))

    operator, position = value
    return (
        ShapeT.OPERATOR()(shapeable(operator, name=f"{name}[0]"), name=name),
        ScalarT.INT().ge(0)(position, name=f"{name}[1]"),
    )


_embed_validator = type_pipe([_embed_tuple, sequence_like(_embed_tuple)], _embed_messenger)


class TensorGraph:
    """
    Base class implementing Tensor methods.
    """

    @validated
    def tensor(
        self,
        data: Annotated[Union[int, float, complex, np.ndarray, Tensor], pipe(scalar_or_shapeable)],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Create a real or complex Tensor with the data provided.

        Parameters
        ----------
        data : number or np.ndarray or Tensor
            The data to convert to an appropriate tensor.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            Real or complex Tensor representation of the input data.

        Notes
        -----
        Use this node to create a Tensor from some numeric `data`. Note that you
        can pass numbers or NumPy arrays to operations that accept Tensors.
        """
        operation = create_operation(self.tensor, locals())
        shape = getattr(data, "shape", ())
        return Tensor(operation, shape=shape)

    @validated
    def concatenate(
        self,
        tensors: Annotated[
            List[Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=no_scalar)]],
            Field(min_length=2),
        ],
        axis: Annotated[int, pipe(ScalarT.INT())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Concatenate a list of tensors along a specified dimension.

        Parameters
        ----------
        tensors : list[np.ndarray or Tensor]
            The list of tensors that you want to concatenate. All of them must have the
            same shape in all dimensions except `axis`. You must pass at least two
            elements in this list.
        axis : int
            The dimension along which you want to concatenate the tensors.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The concatenated tensor.

        Notes
        -----
        This node only concatenates on an existing axis, it does not create new
        axes. If you want to stack along a new axis or concatenate scalars, add
        a new axis to the tensors with ``[None]``.

        Examples
        --------
        >>> x = np.array([[1, 2, 3], [4, 5, 6]])
        >>> y = np.array([[7, 8, 9]])

        Concatenate `x` and `y` along their first dimension.

        >>> graph.concatenate(tensors=[x, y], axis=0, name="node_0")
        <Tensor: name="node_0", operation_name="concatenate", shape=(3, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="node_0")
        >>> result["output"]["node_0"]["value"]
        array([[1., 2., 3.],
               [4., 5., 6.],
               [7., 8., 9.]])

        Concatenate two `x` arrays along their second dimension.

        >>> graph.concatenate(tensors=[x, x], axis=1, name="node_1")
        <Tensor: name="node_1", operation_name="concatenate", shape=(2, 6)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="node_1")
        >>> result["output"]["node_1"]["value"]
        array([[1., 2., 3., 1., 2., 3.],
               [4., 5., 6., 4., 5., 6.]])
        """

        Checker.VALUE(
            all(-len(tensor.shape) <= axis < len(tensor.shape) for tensor in tensors),
            "The axis must be a valid dimension of all the tensors.",
            {"axis": axis},
        )

        _axis = axis if axis >= 0 else axis + len(tensors[0].shape)

        lead_shape = tensors[0].shape[:_axis]
        trail_shape = tensors[0].shape[_axis + 1 :]

        Checker.VALUE(
            all(
                tensor.shape[:_axis] == lead_shape and tensor.shape[_axis + 1 :] == trail_shape
                for tensor in tensors
            ),
            "All tensors must have the same size in every dimension, except for the"
            " dimension that corresponds to the value of axis.",
            {"axis": axis},
        )

        shape = lead_shape + (sum(tensor.shape[_axis] for tensor in tensors),) + trail_shape

        operation = create_operation(self.concatenate, locals())
        return Tensor(operation, shape=shape)

    @validated
    def sum(
        self,
        input_tensor: Union[List[Tensor], Annotated[Union[np.ndarray, Tensor], pipe(shapeable)]],
        axis: Optional[
            Annotated[Union[int, List[int]], pipe(sequence_like(ScalarT.INT(), min_length=0))]
        ] = None,
        keepdims: bool = False,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Sum the elements in a tensor (or a list of tensors with the same shape) along
        one or multiple of its axes.

        Parameters
        ----------
        input_tensor : np.ndarray or Tensor or list[Tensor]
            The tensor whose elements you want to sum. If you pass a list of tensors, they must
            all have the same shape, and are interpreted as being stacked along a new first
            dimension (for example, if you pass two 2D tensors of shape ``(3, 4)``, the result
            is equivalent to passing the stacked 3D tensor of shape ``(2, 3, 4)``).
        axis : int or list[int] or None, optional
            The dimension or dimensions along which you want to sum the tensor. Defaults to None, in
            which case this node sums along all axes of the tensor.
        keepdims : bool, optional
            Whether or not to retain summed axes in the output tensor. If True, each dimension in
            `axis` has size 1 in the result; otherwise, the dimensions in `axis` are removed
            from the result. Defaults to False.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The tensor obtained by summing the input tensor along the specified axes (or, if
            `axis` was None, the tensor obtained by summing the input tensor along all of
            the specified axes).

        See Also
        --------
        Graph.einsum : Tensor contraction via Einstein summation convention.

        Examples
        --------
        Sum elements of an array.

        >>> a = np.array([1, 2, 3])
        >>> graph.sum(a, 0, name="sum_a")
        <Tensor: name="sum_a", operation_name="sum", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="sum_a")
        >>> result["output"]["sum_a"]["value"]
        6

        Sum elements of a 2D array along its first dimension.

        >>> b = np.array([[1, 2, 3], [4, 5, 6]])
        >>> graph.sum(b, 0, name="sum_b")
        <Tensor: name="sum_b", operation_name="sum", shape=(3,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="sum_b")
        >>> result["output"]["sum_b"]["value"]
        array([5, 7, 9])
        """
        if isinstance(input_tensor, list):
            if len(input_tensor) > 0:
                Checker.VALUE(
                    all(
                        tensor.shape == input_tensor[0].shape
                        for tensor in input_tensor[1:]  # type: ignore
                    ),
                    "The tensors must have the same shape.",
                )
                shape = (len(input_tensor),) + input_tensor[0].shape
            else:
                # Note that if the input is an empty list then the shape is somewhat ambiguous (it
                # could be an empty list of tensors of any shape), but for consistency with
                # TF and NP we interpret it as 1D.
                shape = ()
        else:
            shape = input_tensor.shape

        axes = check_operation_axis(axis, shape, "input_tensor")

        output_shape = get_keepdims_operation_shape(shape, axes, keepdims)
        operation = create_operation(self.sum, locals())
        return Tensor(operation, shape=output_shape)

    @validated
    def reverse(
        self,
        tensor: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=no_scalar)],
        axis: Annotated[Union[int, List[int]], pipe(sequence_like(ScalarT.INT()))],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Reverse a tensor along some specified dimensions.

        Parameters
        ----------
        tensor : np.ndarray or Tensor
            The tensor that you want to reverse. It must have at least
            one dimension.
        axis : int or list[int]
            The dimension or dimensions along which you want to reverse the tensor.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The reversed tensor.

        Examples
        --------
        >>> x = np.array([[1, 2, 3], [4, 5, 6]])

        Reverse an array along its first dimension.

        >>> graph.reverse(x, 0, name="a")
        <Tensor: name="a", operation_name="reverse", shape=(2, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="a")
        >>> result["output"]["a"]["value"]
        array([[4, 5, 6],
               [1, 2, 3]])

        Reverse an array along its first and second dimension.

        >>> graph.reverse(x, [0, 1], name="b")
        <Tensor: name="b", operation_name="reverse", shape=(2, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="b")
        >>> result["output"]["b"]["value"]
        array([[6, 5, 4],
               [3, 2, 1]])
        """
        shape = tensor.shape
        _ = check_operation_axis(axis, shape, "tensor")

        operation = create_operation(self.reverse, locals())
        return Tensor(operation, shape=shape)

    @validated
    def repeat(
        self,
        input: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=no_scalar)],
        repeats: Annotated[
            Union[int, List[int]],
            pipe(sequence_like(ScalarT.INT().ge(0), normalizer=list)),
        ],
        axis: Optional[Annotated[int, pipe(ScalarT.INT())]] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Repeat elements of a tensor.

        Parameters
        ----------
        input : np.ndarray or Tensor
            The tensor whose elements you want to repeat.
        repeats : int or list[int]
            The number of times to repeat each element. If you pass a single int or singleton list,
            that number of repetitions is applied to each element. Otherwise, you must pass a list
            with the same length as `input` along the specified `axis` (or the same total length as
            `input` if you omit `axis`).
        axis : int or None, optional
            The axis along which you want to repeat elements. If you omit this value then the input
            is first flattened, and the repetitions applied to the flattened tensor.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The repeated tensor. The result has the same shape as `input` except along `axis`,
            where its size is either the sum of `repeats` (if `repeats` is a list with at least
            two elements) or the product of the original size along `axis` with `repeats` (if
            `repeats` is a single int or singleton list). If `axis` is None then the output is
            1D, with the sizes as described above applied to the flattened input tensor.

        Examples
        --------
        >>> x = np.array([1, 2, 3])
        >>> y = np.array([[1, 2, 3], [4, 5, 6]])

        Duplicate each entry in an array once.

        >>> graph.repeat(x, 2, axis=0, name="a")
        <Tensor: name="a", operation_name="repeat", shape=(6,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="a")
        >>> result["output"]["a"]["value"]
        array([1, 1, 2, 2, 3, 3])

        Create a new array with different repetitions for each element in the original array
        along its second dimension.

        >>> graph.repeat(x, [2, 3, 4], axis=0, name="b")
        <Tensor: name="b", operation_name="repeat", shape=(9,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="b")
        >>> result["output"]["b"]["value"]
        array([1, 1, 2, 2, 2, 3, 3, 3, 3])

        Duplicate each entry in an array along its second dimension.

        >>> graph.repeat(y, 2, axis=1, name="c")
        <Tensor: name="c", operation_name="repeat", shape=(2, 6)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="c")
        >>> result["output"]["c"]["value"]
        array([[1, 1, 2, 2, 3, 3],
               [4, 4, 5, 5, 6, 6]])

        Create a new array with different repetitions for each element in the original array
        along its first dimension.

        >>> graph.repeat(y, [2, 3], axis=0, name="d")
        <Tensor: name="d", operation_name="repeat", shape=(5, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="d")
        >>> result["output"]["d"]["value"]
        array([[1, 2, 3],
               [1, 2, 3],
               [4, 5, 6],
               [4, 5, 6],
               [4, 5, 6]])
        """
        assert isinstance(repeats, list)
        # Prevent the creation of empty tensors.
        Checker.VALUE(
            sum(repeats) > 0,
            "At least one of the repeats must be different from zero.",
            {"repeats": repeats},
        )

        if axis is None:
            # The input is flattened if axis is None.
            shape: Tuple[int, ...] = (np.prod(input.shape, dtype=int),)
            _axis = 0
            _error_variable = "number of elements in input"
        else:
            shape = input.shape
            _axis = check_operation_axis(axis, shape, "input")[0]
            _error_variable = "length of the input along the axis"

        Checker.VALUE(
            len(repeats) == 1 or len(repeats) == shape[_axis],
            f"Number of repeats must either be one or equal to the {_error_variable}.",
            {"len(repeats)": len(repeats), _error_variable: shape[_axis]},
        )

        _repeats = repeats if len(repeats) > 1 else repeats * shape[_axis]

        output_shape = shape[:_axis] + (sum(_repeats),) + shape[_axis + 1 :]
        operation = create_operation(self.repeat, locals())
        return Tensor(operation, shape=output_shape)

    @validated
    def cumulative_sum(
        self,
        x: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=no_scalar)],
        axis: Annotated[int, pipe(ScalarT.INT())] = 0,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Calculate the cumulative sum of a tensor along a specified dimension.

        Parameters
        ----------
        x : np.ndarray or Tensor
            The tensor whose elements you want to sum. It must have at least
            one dimension.
        axis : int, optional
            The dimension along which you want to sum the tensor. Defaults to 0.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The cumulative sum of `x` along dimension `axis`.

        Examples
        --------
        >>> x = np.array([1, 2, 3])
        >>> y = np.array([[1, 2, 3], [4, 5, 6]])

        Calculate the cumulative sum of an array.

        >>> graph.cumulative_sum(x, axis=0, name="a")
        <Tensor: name="a", operation_name="cumulative_sum", shape=(3,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="a")
        >>> result["output"]["a"]["value"]
        array([1, 3, 6])

        Calculate the cumulative sum of a 2D array along its first dimension.

        >>> graph.cumulative_sum(y, axis=0, name="b")
        <Tensor: name="b", operation_name="cumulative_sum", shape=(2, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="b")
        >>> result["output"]["b"]["value"]
        array([[1, 2, 3],
               [5, 7, 9]])

        Calculate the cumulative sum of a 2D array along its second dimension.

        >>> graph.cumulative_sum(y, axis=1, name="c")
        <Tensor: name="c", operation_name="cumulative_sum", shape=(2, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="c")
        >>> result["output"]["c"]["value"]
        array([[ 1,  3,  6],
               [ 4,  9, 15]])
        """
        shape = x.shape
        _ = check_operation_axis(axis, shape, "x")

        operation = create_operation(self.cumulative_sum, locals())
        return Tensor(operation, shape=shape)

    @validated
    def transpose(
        self,
        a: Annotated[Union[np.ndarray, Tensor], pipe(shapeable)],
        perm: Optional[List[Annotated[int, pipe(ScalarT.INT())]]] = None,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Reorder the dimensions of a tensor.

        Parameters
        ----------
        a : np.ndarray or Tensor
            The tensor whose dimensions you want to permute, :math:`x`.
        perm : list[int] or None, optional
            The order of the input dimensions for the returned tensor. If you provide it, it must
            be a permutation of all integers between 0 and ``N-1``, where `N` is the rank of `a`.
            If you don't provide it, the order of the dimensions is inverted, that is to say,
            it defaults to ``[N-1, …, 1, 0]``.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The input tensor with its dimensions permuted. The `i`-th dimension of the
            returned tensor corresponds to the `perm[i]`-th input dimension.

        See Also
        --------
        Graph.adjoint : Hermitian adjoint of an operator.
        Graph.einsum : Tensor contraction via Einstein summation convention.
        """

        if perm is None:
            shape = a.shape[::-1]
        else:
            # This check converts negative axes into the range [0, len(a.shape)]:
            axes = check_operation_axis(perm, a.shape, "a")

            # This check makes sure all axes in the range [0, len(a.shape)] were used:
            Checker.VALUE(
                set(axes) == set(range(len(a.shape))),
                "The value of perm must be a valid permutation of the indices of a.",
                {"perm": perm, "missing indices": set(range(len(a.shape))) - set(axes)},
            )

            shape = tuple(a.shape[dimension] for dimension in axes)

        operation = create_operation(self.transpose, locals())
        return Tensor(operation, shape=shape)

    @validated
    def pauli_matrix(
        self,
        label: Literal["I", "X", "Y", "Z", "M", "P"],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Create a Pauli matrix from a label.

        Parameters
        ----------
        label : str
            The string that indicates which Pauli matrix to create.
            Must be ``'I'``, ``'X'``, ``'Y'``, ``'Z'``, ``'M'``, or ``'P'``. ``'M'`` creates
            the lowering matrix :math:`\sigma_- = \frac{1}{2}(\sigma_x + i\sigma_y)`. ``'P'``
            creates the raising matrix :math:`\sigma_+ = \frac{1}{2}(\sigma_x - i\sigma_y)`.
            We use the convention :math:`|\downarrow\rangle = \begin{bmatrix}1\\0\end{bmatrix}`
            and :math:`|\uparrow\rangle = \begin{bmatrix}0\\1\end{bmatrix}`.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The Pauli matrix.

        See Also
        --------
        Graph.pauli_kronecker_product : Embed Pauli matrices into a larger Hilbert space.

        Examples
        --------
        Create the Pauli X matrix.

        >>> graph.pauli_matrix("X", name="sigma_x")
        <Tensor: name="sigma_x", operation_name="pauli_matrix", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="sigma_x")
        >>> result["output"]["sigma_x"]["value"]
        array([[0.+0.j, 1.+0.j],
               [1.+0.j, 0.+0.j]])
        """
        operation = create_operation(self.pauli_matrix, locals())
        return Tensor(operation, shape=(2, 2))

    @validated
    def pauli_kronecker_product(
        self,
        labels: Annotated[
            List[
                Tuple[
                    Literal["I", "X", "Y", "Z", "M", "P"],
                    Annotated[int, pipe(ScalarT.INT().ge(0))],
                ]
            ],
            Field(min_length=1),
        ],
        subsystem_count: Annotated[int, pipe(ScalarT.INT().gt(0))],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Place Pauli matrices into their two-dimensional subsystems of a system and
        returns the Kronecker product.

        Parameters
        ----------
        labels : list[tuple[str, int]]
            A list of tuples, each containing a pair of labels for the Pauli matrix and its
            position.  The Pauli matrix label is a string ``'I'``, ``'X'``, ``'Y'``, ``'Z'``,
            ``'M'``, or ``'P'`` and the position label is a non-negative integer and smaller
            than `system_count` indicating the position of the Pauli matrix in the system. At
            least one tuple must be provided.  ``'M'`` creates the lowering matrix
            :math:`\sigma_- = \frac{1}{2}(\sigma_x + i\sigma_y)`.  ``'P'`` creates the raising
            matrix :math:`\sigma_+ = \frac{1}{2}(\sigma_x - i\sigma_y)`.
            We use the convention :math:`|\downarrow\rangle = \begin{bmatrix}1\\0\end{bmatrix}`
            and :math:`|\uparrow\rangle = \begin{bmatrix}0\\1\end{bmatrix}`.
        subsystem_count : int
            The number of two-level subsystems that constitute the system. Must be a positive
            number.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The Kronecker product of Pauli matrices.

        See Also
        --------
        Graph.embed_operators : Embed operators into a larger Hilbert space.
        Graph.kron : Kronecker product between two objects.
        Graph.kronecker_product_list : Kronecker product of a list of operators.

        Examples
        --------
        Place a single Pauli :math:`X` matrix in the second of two subsystems to create :math:`IX`.

        >>> graph.pauli_kronecker_product([("X", 1)], subsystem_count=2, name="IX")
        <Tensor: name="IX", operation_name="pauli_kronecker_product", shape=(4, 4)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="IX")
        >>> result["output"]["IX"]["value"]
        array([[0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
               [1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j],
               [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j]])

        Place a single Pauli :math:`X` matrix in the second of three subsystems to create
        :math:`IXI`.

        >>> graph.pauli_kronecker_product([("X", 1)], subsystem_count=3, name="IXI")
        <Tensor: name="IXI", operation_name="pauli_kronecker_product", shape=(8, 8)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="IXI")
        >>> result["output"]["IXI"]["value"]
        array([[0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               ...
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j]])

        Place two Pauli :math:`X` matrices in the second and third of three subsystems
        to create :math:`IXX`.

        >>> graph.pauli_kronecker_product([("X", 1), ("X", 2)], subsystem_count=3, name="IXX")
        <Tensor: name="IXX", operation_name="pauli_kronecker_product", shape=(8, 8)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="IXX")
        >>> result["output"]["IXX"]["value"]
        array([[0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               ...
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j]])
        """
        positions = [label[1] for label in labels]

        Checker.VALUE(
            np.all(np.array(positions) < subsystem_count),
            "All positions must be smaller than `subsystem_count`.",
            {"positions": positions, "subsystem_count": subsystem_count},
        )
        Checker.VALUE(len(set(positions)) == len(positions), "All positions must be unique.")

        dimension = 2**subsystem_count
        operation = create_operation(self.pauli_kronecker_product, locals())
        return Tensor(operation, shape=(dimension, dimension))

    @validated
    def einsum(
        self,
        equation: str,
        tensors: List[Annotated[Union[np.ndarray, Tensor], pipe(shapeable)]],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Perform tensor contraction via Einstein summation convention.

        Use this node to perform multi-dimensional, linear algebraic array operations between
        tensors.

        Parameters
        ----------
        equation : str
            The equation describing the tensor contraction.
            The format is the same as in NumPy's einsum.
        tensors : list[np.ndarray or Tensor]
            The tensors to be contracted. Their rank must be not greater than six.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The contracted Tensor.

        See Also
        --------
        Graph.matmul : Matrix multiplication between objects.
        Graph.sum : Sum elements in a tensor along one or multiple axes.
        Graph.trace : Trace of an object.
        Graph.transpose : Reorder the dimensions of a tensor.

        Notes
        -----
        You can use tensor contraction with Einstein summation convention to create a new
        tensor from its element-wise computation from other tensors. This applies to any
        tensor operation that you can write as an equation relating the elements of the
        result as sums over products of elements of the inputs.

        The element-wise equation of the operation is summarized by a string describing
        the Einstein summation to be performed on the inputs. For example, the matrix
        multiplication between two matrices can be written as

        .. math::
            R_{ik} = \sum_j P_{ij} Q_{jk} .

        To convert this element-wise equation to the appropriate string, you can:
        remove summations and variable names (`ik = ij * jk`),
        move the output to the right (`ij * jk = ik`), and
        replace "`*`" with "`,`" and "`=`" with "`->`" (`ij,jk->ik`).
        You can also use an ellipsis (...) to broadcast over unchanged dimensions.

        For more information about Einstein summation, see `Einstein notation`_ on Wikipedia.

        .. _Einstein notation:
            https://en.wikipedia.org/wiki/Einstein_notation

        Examples
        --------
        >>> x = np.arange(16, dtype=float)

        Diagonal of a matrix.

        >>> graph.einsum("ii->i", [x.reshape(4, 4)], name="diagonal")
        <Tensor: name="diagonal", operation_name="einsum", shape=(4,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="diagonal")
        >>> result["output"]["diagonal"]["value"]
        array([0., 5., 10., 15.])

        Trace of a matrix.

        >>> graph.einsum('ii->', [x.reshape(4, 4)], name="trace")
        <Tensor: name="trace", operation_name="einsum", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="trace")
        >>> result["output"]["trace"]["value"]
        30.0

        Sum over matrix axis.

        >>> graph.einsum('ij->i', [x.reshape((4, 4))], name="sum_1")
        <Tensor: name="sum_1", operation_name="einsum", shape=(4,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="sum_1")
        >>> result["output"]["sum_1"]["value"]
        array([ 6., 22., 38., 54.])

        Sum over tensor axis ignoring leading dimensions.

        >>> graph.einsum('...ji->...i', [x.reshape((2, 2, 4))], name='sum_2')
        <Tensor: name="sum_2", operation_name="einsum", shape=(2, 4)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="sum_2")
        >>> result["output"]["sum_2"]["value"]
        array([[ 4.,  6.,  8., 10.],
               [20., 22., 24., 26.]])

        Reorder tensor axes.

        >>> graph.einsum('ijk->jki', [x.reshape((8, 1, 2))], name="reorder")
        <Tensor: name="reorder", operation_name="einsum", shape=(1, 2, 8)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="reorder")
        >>> result["output"]["reorder"]["value"]
        array([[[ 0.,  2.,  4.,  6.,  8., 10., 12., 14.],
                [ 1.,  3.,  5.,  7.,  9., 11., 13., 15.]]])

        Vector inner product.

        >>> graph.einsum('i,i->', [x, np.ones(16)], name="inner")
        <Tensor: name="inner", operation_name="einsum", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="inner")
        >>> result["output"]["inner"]["value"]
        120.0

        Matrix-vector multiplication.

        >>> graph.einsum('ij,j->i', [x.reshape((4, 4)), np.ones(4)], name="multiplication")
        <Tensor: name="multiplication", operation_name="einsum", shape=(4,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="multiplication")
        >>> result["output"]["multiplication"]["value"]
        array([ 6., 22., 38., 54.])

        Vector outer product.

        >>> graph.einsum("i,j->ij", [x[:2], x[:3]], name="outer")
        <Tensor: name="outer", operation_name="einsum", shape=(2, 3)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="outer")
        >>> result["output"]["outer"]["value"]
        array([[0., 0., 0.],
               [0., 1., 2.]])

        Tensor contraction.

        >>> graph.einsum(
        ...     "ijk,jil->kl", [x.reshape((4, 2, 2)), x.reshape((2, 4, 2))], name="contraction"
        ... )
        <Tensor: name="contraction", operation_name="einsum", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="contraction")
        >>> result["output"]["contraction"]["value"]
        array([[504., 560.],
               [560., 624.]])

        Trace along first two axes.

        >>> graph.einsum("ii...->i...", [x.reshape((2, 2, 4))], name="trace_2")
        <Tensor: name="trace_2", operation_name="einsum", shape=(2, 4)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="trace_2")
        >>> result["output"]["trace_2"]["value"]
        array([[ 0.,  1.,  2.,  3.],
               [12., 13., 14., 15.]])

        Matrix multiplication using the left-most indices.

        >>> graph.einsum(
        ...     "ij...,jk...->ik...", [x.reshape((1, 4, 4)), x.reshape((4, 1, 4))], name="left_most"
        ... )
        <Tensor: name="left_most", operation_name="einsum", shape=(1, 1, 4)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="left_most")
        >>> result["output"]["left_most"]["value"]
        array([[[224., 276., 336., 404.]]])
        """

        try:
            shape = np.einsum(equation, *[np.zeros(tensor.shape) for tensor in tensors]).shape
        except ValueError as exc:
            raise ValueError(
                "The equation is not valid or is incompatible with the inputs.",
            ) from exc

        Checker.VALUE(
            all(len(tensor.shape) <= 6 for tensor in tensors),
            "Tensors with rank greater than six are not supported.",
        )
        operation = create_operation(self.einsum, locals())
        return Tensor(operation, shape=shape)

    @validated
    def expectation_value(
        self,
        state: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.VECTOR())],
        operator: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.OPERATOR())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the expectation value of an operator with respect to a state.

        The last dimension of the state must be equal to the last two dimensions
        of the operator and their batch shapes must be broadcastable.

        Parameters
        ----------
        state : np.ndarray or Tensor
            The state. It must be a vector of shape ``(..., D)``.
        operator : np.ndarray or Tensor
            The operator. It must be of shape ``(..., D, D)``.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The expectation value with shape ``(...)``.

        See Also
        --------
        Graph.density_matrix_expectation_value :
            Expectation value of an operator with respect to a density matrix.
        Graph.inner_product : Inner product of two vectors.
        Graph.outer_product : Outer product of two vectors.
        Graph.trace : Trace of an object.

        Notes
        -----
        The expectation value of an operator :math:`\mathbf{A}` with respect to
        a vector :math:`\mathbf{x}` is defined as

        .. math::
            \mathbf{x}^\dagger \mathbf{A} \mathbf{x} = \langle x \vert \mathbf{A} \vert x \rangle
            = \sum_{ij} x^\ast_{i} A_{ij} x_{j} .

        For more information about the expectation value, see `expectation value`_ on Wikipedia.

        .. _expectation value:
            https://en.wikipedia.org/wiki/Expectation_value_(quantum_mechanics)

        Examples
        --------
        >>> graph.expectation_value(np.array([1j, 1j]), np.eye(2), name="expectation")
        <Tensor: name="expectation", operation_name="expectation_value", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="expectation")
        >>> result["output"]["expectation"]["value"]
        2.+0.j
        >>> graph.expectation_value(np.ones([3,1,4]), np.ones([2,4,4]), name="expectation)
        <Tensor: name="expectation", operation_name="expectation_value", shape=(3, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="expectation")
        >>> result["output"]["expectation"]["value"]
        array([[16, 16], [16, 16], [16, 16]])
        """

        Checker.VALUE(
            state.shape[-1] == operator.shape[-1],
            "State and operator shapes must have the same size in their last axis.",
            {
                "value shape of state": state.shape[-1:],
                "value shape of operator": operator.shape[-2:],
            },
        )

        shape = get_broadcasted_shape(
            state.shape[:-1],
            operator.shape[:-2],
            message="The batch shapes of state and operator must be broadcastable.",
        )

        operation = create_operation(self.expectation_value, locals())
        return Tensor(operation, shape=shape)

    @validated
    def density_matrix_expectation_value(
        self,
        density_matrix: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.OPERATOR()),
        ],
        operator: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.OPERATOR())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the expectation value of an operator with respect to a density matrix.

        The last two dimensions of the density matrix must be equal to the last two dimensions
        of the operator and their batch shapes must be broadcastable.

        Parameters
        ----------
        density_matrix : np.ndarray or Tensor
            The density matrix. It must be of shape ``(..., D, D)``.
        operator : np.ndarray or Tensor
            The operator. It must be of shape ``(..., D, D)``.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The expectation value with shape ``(...)``.

        See Also
        --------
        Graph.expectation_value : Expectation value of an operator with respect to a pure state.
        Graph.inner_product : Inner product of two vectors.
        Graph.outer_product : Outer product of two vectors.
        Graph.trace : Trace of an object.

        Notes
        -----
        The expectation value of an operator :math:`\mathbf{A}` with respect to
        a density matrix :math:`\rho=\sum_i p_i |\psi_i\rangle\langle\psi_i|` is defined as

        .. math::
            {\mathrm{Tr}}(A\rho) = {\mathrm{Tr}}(A\sum_i p_i |\psi_i\rangle\langle\psi_i|)
            = \sum_i p_i \langle\psi_i|A|\psi_i\rangle .

        For more information about the density matrix expectation value, see
        `density matrix`_ on Wikipedia.

        .. _density matrix:
            https://en.wikipedia.org/wiki/Density_matrix

        Examples
        --------
        >>> graph.density_matrix_expectation_value(
        ...     np.array([[0.9, 0.], [0., 0.1]]), np.array([[1., 0.], [0., -1.]]),
        ...     name="expectation",
        ... )
        <Tensor: name="expectation", operation_name="density_matrix_expectation_value", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="expectation")
        >>> result["output"]["expectation"]["value"]
        0.8
        >>> graph.density_matrix_expectation_value(
        ...     np.array([[0.9, 0.], [0., 0.1]]),
        ...     np.array([[[0., 1.], [1., 0.]], [[0., -1.j], [1.j, 0.]], [[1., 0.], [0., -1.]]]),
        ...     name="expectation2"
        ... )
        <Tensor: name="expectation2", operation_name="expectation_value", shape=(3,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="expectation2")
        >>> result["output"]["expectation2"]["value"]
        array([0. +0.j, 0. +0.j, 0.8+0.j])
        """

        Checker.VALUE(
            density_matrix.shape[-1] == operator.shape[-1],
            "Density matrix and operator shapes must have the same size in their last axis.",
            {
                "value shape of density matrix": density_matrix.shape[-2:],
                "value shape of operator": operator.shape[-2:],
            },
        )

        shape = get_broadcasted_shape(
            density_matrix.shape[:-2],
            operator.shape[:-2],
            message="The batch shapes of density matrix and operator must be broadcastable.",
        )

        operation = create_operation(self.density_matrix_expectation_value, locals())
        return Tensor(operation, shape=shape)

    @validated
    def inner_product(
        self,
        x: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.VECTOR())],
        y: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.VECTOR())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the inner product of two vectors.

        The vectors must have the same last dimension and broadcastable shapes.

        Parameters
        ----------
        x : np.ndarray or Tensor
            The left multiplicand. It must be a vector of shape ``(..., D)``.
        y : np.ndarray or Tensor
            The right multiplicand. It must be a vector of shape ``(..., D)``.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The inner product of two vectors of shape ``(...)``.

        See Also
        --------
        Graph.density_matrix_expectation_value :
            Expectation value of an operator with respect to a density matrix.
        Graph.einsum : Tensor contraction via Einstein summation convention.
        Graph.expectation_value : Expectation value of an operator with respect to a pure state.
        Graph.outer_product : Outer product of two vectors.
        Graph.trace : Trace of an object.

        Notes
        -----
        The inner product or dot product of two complex vectors :math:`\mathbf{x}`
        and :math:`\mathbf{y}` is defined as

        .. math::
            \langle \mathbf{x} \vert \mathbf{y} \rangle = \sum_i x^\ast_{i} y_{i} .

        For more information about the inner product, see `dot product`_ on Wikipedia.

        .. _dot product:
            https://en.wikipedia.org/wiki/Dot_product

        Examples
        --------
        >>> graph.inner_product(np.array([1j, 1j]), np.array([1j, 1j]), name="inner")
        <Tensor: name="inner", operation_name="inner_product", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="inner")
        >>> result["output"]["inner"]["value"]
        2.+0.j

        >>> graph.inner_product(np.ones((3,1,4), np.ones(2,4), name="inner2")
        <Tensor: name="inner2", operation_name="inner_product", shape=(3, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="inner2")
        >>> result["output"]["inner2"]["value"]
        array([[4, 4], [4, 4], [4, 4]])
        """

        Checker.VALUE(
            x.shape[-1] == y.shape[-1],
            "The vectors must be of same length in their last dimension.",
            {"x.shape": y.shape, "y.shape": x.shape},
        )

        shape = get_broadcasted_shape(
            x.shape[:-1],
            y.shape[:-1],
            message="The batch shapes of x and y must be broadcastable.",
        )

        operation = create_operation(self.inner_product, locals())
        return Tensor(operation, shape=shape)

    @validated
    def outer_product(
        self,
        x: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.VECTOR())],
        y: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.VECTOR())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the outer product of two vectors.

        The vectors can have different last dimensions but must have broadcastable batch dimensions.

        Parameters
        ----------
        x : np.ndarray or Tensor
            The left multiplicand. It must be a vector of shape ``(..., M)``.
        y : np.ndarray or Tensor
            The right multiplicand. It must be a vector of shape ``(..., N)``.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The outer product of two vectors of shape ``(..., M, N)``.

        See Also
        --------
        Graph.density_matrix_expectation_value :
            Expectation value of an operator with respect to a density matrix.
        Graph.expectation_value : Expectation value of an operator with respect to a pure state.
        Graph.inner_product : Inner product of two vectors.
        Graph.trace : Trace of an object.

        Notes
        -----
        The outer product of two complex vectors :math:`\mathbf{x}`
        and :math:`\mathbf{y}` is defined as

        .. math::
            (\mathbf{x} \otimes \mathbf{y})_{ij} = x_{i} y^\ast_{j}.

        For more information about the outer product, see `outer product`_ on Wikipedia.

        .. _outer product:
            https://en.wikipedia.org/wiki/Outer_product

        Examples
        --------
        >>> graph.outer_product(np.array([1j, 1j]), np.array([1j, -1j]), name="outer")
        <Tensor: name="outer", operation_name="outer_product", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="outer")
        >>> result["output"]["outer"]["value"]
        array([[1.+0.j, -1.+0.j], [1.+0.j, -1.+0.j]])

        >>> graph.outer_product(np.ones((3,1,2), np.ones(2,2), name="outer2")
        <Tensor: name="outer2", operation_name="outer_product", shape=(3, 2, 2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="outer2")
        >>> result["output"]["outer2"]["value"]
        array([[[[1, 1], [1, 1]], [[1, 1], [1, 1]]],
            [[[1, 1], [1, 1]], [[1, 1], [1, 1]]],
            [[[1, 1], [1, 1]], [[1, 1], [1, 1]]])
        """

        batch_shape = get_broadcasted_shape(
            x.shape[:-1],
            y.shape[:-1],
            message="The batch shapes of x and y must be broadcastable.",
        )
        shape = batch_shape + (x.shape[-1], y.shape[-1])

        operation = create_operation(self.outer_product, locals())
        return Tensor(operation, shape=shape)

    @validated
    def partial_trace(
        self,
        density_matrix: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.OPERATOR()),
        ],
        subsystem_dimensions: Annotated[
            List[Annotated[int, pipe(ScalarT.INT().gt(0))]],
            Field(min_length=1),
        ],
        traced_subsystems: Union[
            Annotated[List[Annotated[int, pipe(ScalarT.INT().ge(0))]], Field(min_length=1)],
            Annotated[int, pipe(ScalarT.INT().ge(0))],
        ],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the partial trace of a density matrix.

        Parameters
        ----------
        density_matrix : np.ndarray or Tensor
            The density matrix :math:`\rho` of the system to be reduced.
            Can be a single square matrix or a batch of matrices
            with dimension ``(..., D, D)``.
        subsystem_dimensions : list[int]
            The dimension of each subsystem. The product of the subsystem
            dimensions is the dimension of the system ``D``.
        traced_subsystems : int or list[int]
            The indices (starting from zero) of the subsystems to be traced out.
            Each index refers to a different subsystem.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The reduced density matrix of shape ``(..., d, d)``. The reduced
            dimension ``d`` is equal to the system dimension ``D`` divided by the
            product of the traced out subsystem dimensions.

        Notes
        -----
        Given a density matrix :math:`\rho` of two subsystems :math:`A`
        and :math:`B`, the partial trace over subsystem :math:`B` is defined as

        .. math::
            ({\mathrm{Tr}_{B}} \rho)_{ij} = \sum_k \rho_{ik,jk}.

        For more information about the partial trace, see `partial trace`_ on Wikipedia.

        .. _partial trace:
            https://en.wikipedia.org/wiki/Partial_trace

        Examples
        --------
        >>> graph.partial_trace(np.diag([1, 0, 0, 0]), [2, 2], 1, name="partial")
        <Tensor: name="partial", operation_name="partial_trace", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="partial")
        >>> result["output"]["partial"]["value"]
        array([[[1, 0], [0, 0]])

        >>> graph.partial_trace(np.eye(10)/10, [2, 5], 1, name="partial2")
        <Tensor: name="partial2", operation_name="partial_trace", shape=(2, 2)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="partial2")
        >>> result["output"]["partial2"]["value"]
        array([[[0.5, 0], [0, 0.5]])

        See more examples in the `How to simulate large open system dynamics
        <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-simulate-large-open-
        system-dynamics>`_ user guide.
        """

        Checker.VALUE(
            np.prod(subsystem_dimensions) == density_matrix.shape[-1],
            "The product of the subsystem_dimensions must be equal to the last "
            "density_matrix dimension.",
        )

        if not isinstance(traced_subsystems, list):
            traced_subsystems = [traced_subsystems]

        Checker.VALUE(
            np.all(np.array(traced_subsystems) < len(subsystem_dimensions)),
            "The traced_subsystems must be indices within the range "
            "of the number of subsystems.",
        )
        Checker.VALUE(
            len(set(traced_subsystems)) == len(traced_subsystems),
            "The traced_subsystems must not have duplicate values.",
        )

        reduced_matrix_dimension = density_matrix.shape[-1] // np.prod(
            [subsystem_dimensions[i] for i in traced_subsystems],
            dtype=int,
        )
        shape = density_matrix.shape[:-2] + (
            reduced_matrix_dimension,
            reduced_matrix_dimension,
        )

        operation = create_operation(self.partial_trace, locals())
        return Tensor(operation, shape=shape)

    @validated
    def reshape(
        self,
        tensor: Annotated[Union[np.ndarray, Tensor], pipe(shapeable)],
        shape: Tuple[Annotated[int, pipe(ScalarT.INT().ge(-1).ne(0))], ...],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Reshape a tensor into a new shape, keeping the order of its elements.

        Parameters
        ----------
        tensor : np.ndarray or Tensor
            The tensor you want to reshape.
        shape : tuple[int, ...]
            The new shape of the tensor.
            One and only one of the dimensions can be set to -1.
            In that case, the method will automatically calculate that dimension's size by
            keeping the total size constant.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The reshaped tensor.

        Examples
        --------
        >>> graph.reshape(tensor=np.ones((4, 4)), shape=(2, 8), name='reshape')
        <Tensor: name="reshape", operation_name="reshape", shape=(2, 8)>

        >>> graph.reshape(tensor=np.ones((4, 4)), shape=(2, -1), name='reshape_1')
        <Tensor: name="reshape_1", operation_name="reshape", shape=(2, 8)>

        >>> graph.reshape(tensor=np.ones((4, 4)), shape=(2, -1, 2), name='reshape_2')
        <Tensor: name="reshape_2", operation_name="reshape", shape=(2, 4, 2)>

        >>> graph.reshape(tensor=np.ones((2, 2)), shape=(-1,), name='reshape_3')
        <Tensor: name="reshape_3", operation_name="reshape", shape=(4,)>
        """
        tensor_element_count = np.prod(tensor.shape)
        shape_element_count = np.prod(shape)

        missing_dimensions = np.nonzero(np.asarray(shape) == -1)[0]
        if len(missing_dimensions) > 0:
            Checker.VALUE(
                len(missing_dimensions) == 1,
                "You can only specify one axis with -1 in the new shape.",
                {"shape": shape},
            )

            _dim = missing_dimensions[0]
            missing_value = (-1) * tensor_element_count / shape_element_count

            Checker.VALUE(
                int(missing_value) == missing_value,
                "Unable to allocate a whole number of elements for the unspecified axis (-1).",
                {"tensor.shape": tensor.shape, "shape": shape},
            )
            shape_new = shape[:_dim] + (int(missing_value),) + shape[_dim + 1 :]
        else:
            Checker.VALUE(
                tensor_element_count == shape_element_count,
                "New shape must have the same number of elements as the input tensor.",
                {"tensor.shape": tensor.shape, "shape": shape},
            )
            shape_new = shape

        operation = create_operation(self.reshape, locals())
        return Tensor(operation, shape=shape_new)

    @validated
    def density_matrix_infidelity(
        self,
        x: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.OPERATOR())],
        y: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.OPERATOR())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the infidelity between two states represented by density matrices.

        Parameters
        ----------
        x : np.ndarray or Tensor
            The density matrix :math:`x` with shape ``(..., D, D)``.
            The last two dimensions must have the same size for both
            matrices, and its batch dimensions must be broadcastable.
        y : np.ndarray or Tensor
            The density matrix :math:`y` with shape ``(..., D, D)``.
            The last two dimensions must have the same size for both
            matrices, and its batch dimensions must be broadcastable.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The state infidelity of the density matrices with respect to each
            other. Its shape is the broadcasted value of the batch shapes of
            the two input parameters.

        Warnings
        --------
        This function assumes that the parameters are density matrices and
        therefore are positive definite. Passing matrices that have negative
        or complex eigenvalues will result in wrong values for the infidelity.

        See Also
        --------
        Graph.infidelity_pwc : Total infidelity of a system with a piecewise-constant Hamiltonian.
        Graph.infidelity_stf : Total infidelity of a system with a sampleable Hamiltonian.
        Graph.state_infidelity : Infidelity between two quantum states.
        Graph.unitary_infidelity : Infidelity between a unitary and target operators.

        Notes
        -----
        The general formula for the infidelity of two density matrices is

        .. math::
            I = 1 - \left[ \mathrm{Tr}\left( \sqrt{\sqrt{x} y \sqrt{x}} \right) \right]^2

        Examples
        --------
        >>> infidelity = graph.density_matrix_infidelity(
        ...     np.array([[0.5, 0], [0, 0.5]]),
        ...     np.array([[1, 0], [0, 0]]),
        ...     name="infidelity",
        ... )
        >>> result = bo.execute_graph(graph=graph, output_node_names="infidelity")
        >>> result["output"]["infidelity"]["value"]
        0.5
        """

        Checker.VALUE(
            x.shape[-2:] == y.shape[-2:],
            "The last two dimensions of the x and y must be the same.",
            {"x.shape": x.shape, "y.shape": y.shape},
        )

        broadcasted_shape = get_broadcasted_shape(
            x.shape[:-2],
            y.shape[:-2],
            message="The batch shapes of x and y must be broadcastable.",
        )

        operation = create_operation(self.density_matrix_infidelity, locals())
        return Tensor(operation, shape=broadcasted_shape)

    @validated
    def unitary_infidelity(
        self,
        unitary_operator: Annotated[
            Union[np.ndarray, Tensor],
            pipe(shapeable, after=ShapeT.OPERATOR()),
        ],
        target: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.OPERATOR())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the infidelity between a target operation and the actual implemented unitary.

        Both operators must be square and have shapes broadcastable to each other.

        Parameters
        ----------
        unitary_operator : np.ndarray or Tensor
            The actual unitary operator, :math:`U`, with shape ``(..., D, D)``.
            Its last two dimensions must be equal and the same as `target`, and its batch
            dimensions, if any, must be broadcastable with `target`.
        target : np.ndarray or Tensor
            The target operation with respect to which the infidelity will be calculated,
            :math:`V`, with shape ``(..., D, D)``.
            Its last two dimensions must be equal and the same as `unitary_operator`,
            and its batch dimensions, if any, must be broadcastable with `unitary_operator`.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The infidelity between the two operators, with shape ``(...)``.

        See Also
        --------
        Graph.density_matrix_infidelity : Infidelity between two density matrices.
        Graph.infidelity_pwc : Total infidelity of a system with a piecewise-constant Hamiltonian.
        Graph.infidelity_stf : Total infidelity of a system with a sampleable Hamiltonian.
        Graph.state_infidelity : Infidelity between two quantum states.

        Notes
        -----
        The operational infidelity between the actual unitary and target operators is defined as

        .. math::
          \mathcal{I} = 1 - \left|
              \frac{\mathrm{Tr} (V^\dagger U)}{\mathrm{Tr} (V^\dagger V)}
          \right|^2 .

        Examples
        --------
        Calculate the infidelity of a unitary with respect to a :math:`\sigma_x` gate.

        >>> theta = 0.5
        >>> sigma_x = np.array([[0, 1], [1, 0]])
        >>> unitary = np.array([[np.cos(theta), np.sin(theta)], [np.sin(theta), -np.cos(theta)]])
        >>> graph.unitary_infidelity(unitary_operator=unitary, target=sigma_x, name="infidelity")
        <Tensor: name="infidelity", operation_name="unitary_infidelity", shape=()>
        >>> result = bo.execute_graph(graph=graph, output_node_names="infidelity")
        >>> result["output"]["infidelity"]["value"]
        0.7701511529340699

        Calculate the time-dependent infidelity of the identity gate for a noiseless single qubit.

        >>> sigma_x = np.array([[0, 1], [1, 0]])
        >>> hamiltonian = sigma_x * graph.pwc_signal(
        ...     duration=1, values=np.pi * np.array([0.25, 1, 0.25])
        ... )
        >>> unitaries = graph.time_evolution_operators_pwc(
        ...     hamiltonian=hamiltonian, sample_times=np.linspace(0, 1, 10)
        ... )
        >>> graph.unitary_infidelity(
        ...     unitary_operator=unitaries, target=np.eye(2), name="infidelities"
        ... )
        <Tensor: name="infidelities", operation_name="unitary_infidelity", shape=(10,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="infidelities")
        >>> result["output"]["infidelities"]["value"]
        array([0.        , 0.00759612, 0.03015369, 0.0669873 , 0.32898993,
               0.67101007, 0.9330127 , 0.96984631, 0.99240388, 1.        ])
        """
        Checker.VALUE(
            unitary_operator.shape[-2:] == target.shape[-2:],
            "The last two dimensions of the unitary and target operators must be the same.",
            {
                "unitary_operator.shape": unitary_operator.shape,
                "target.shape": target.shape,
            },
        )

        shape = get_broadcasted_shape(
            unitary_operator.shape[:-2],
            target.shape[:-2],
            message="The batch shapes of unitary operator and target must be broadcastable.",
        )

        operation = create_operation(self.unitary_infidelity, locals())
        return Tensor(operation, shape=shape)

    @validated
    def state_infidelity(
        self,
        x: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.VECTOR())],
        y: Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.VECTOR())],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the infidelity of two pure states.

        Parameters
        ----------
        x : np.ndarray or Tensor
            A pure state, :math:`|\psi\rangle`, with shape ``(..., D)``.
            Note that the last dimension must be the same as `y`, and the batch dimension,
            if any, must be broadcastable with `y`.
        y : np.ndarray or Tensor
            A pure state, :math:`|\phi\rangle`, with shape ``(..., D)``.
            Note that the last dimension must be the same as `x`, and the batch dimension,
            if any, must be broadcastable with `x`.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The infidelity of two pure states, with shape ``(...)``.

        See Also
        --------
        Graph.density_matrix_infidelity : Infidelity between two density matrices.
        Graph.inner_product : Inner product of two vectors.
        Graph.unitary_infidelity : Infidelity between a unitary and target operators.

        Notes
        -----
        The infidelity of two pure states :math:`|\psi\rangle` and :math:`|\phi\rangle`
        is defined as :math:`1 - \| \langle \psi | \phi \rangle \|^2`.

        For more information about the state fidelity, see `fidelity of quantum states`_
        on Wikipedia.

        .. _fidelity of quantum states:
            https://en.wikipedia.org/wiki/Fidelity_of_quantum_states

        Examples
        --------
        >>> graph.state_infidelity(
        ...     np.array([0, 1]), np.array([[1, 0], [0, 1]]), name="infidelity"
        ... )
        <Tensor: name="infidelity", operation_name="state_infidelity", shape=(2,)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="infidelity")
        >>> result["output"]["infidelity"]["value"]
        array([1., 0.])
        """

        Checker.VALUE(
            x.shape[-1] == y.shape[-1],
            "The last dimension of x and y must be the same.",
            {"x.shape": x.shape, "y.shape": y.shape},
        )
        shape = get_broadcasted_shape(
            x.shape[:-1],
            y.shape[:-1],
            message="The batch shapes of x and y must be broadcastable.",
        )

        operation = create_operation(self.state_infidelity, locals())
        return Tensor(operation, shape=shape)

    @validated
    def embed_operators(
        self,
        operators: Annotated[
            Union[
                List[Tuple[Union[np.ndarray, Tensor], int]],
                Tuple[Union[np.ndarray, Tensor], int],
            ],
            pipe(_embed_validator),
        ],
        dimensions: List[Annotated[int, pipe(ScalarT.INT().gt(0))]],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Embed an operator or set of operators into a larger Hilbert space.

        Parameters
        ----------
        operators : list[tuple[np.ndarray or Tensor, int]] or tuple[np.ndarray or Tensor, int]
            A tuple or a list of tuples that contain pairs of the operators to be embedded and their
            corresponding positions in the full space.
            The positions must be non-negative integers less than the length of the dimensions list.
            The operators must be at least 2D and can contain leading batch dimensions.
            The dimension of the operators must match the value of the dimensions list
            at the given position.
        dimensions : list[int]
            The dimensions of each subspace.
        name : str or none, optional
            The name of the node.

        Returns
        -------
        Tensor
            The operators embedded into the full Hilbert space. The dimension of the Hilbert
            space of the output is equal to the product of the elements of the dimensions list.

        See Also
        --------
        Graph.kron : Kronecker product between two objects.
        Graph.kronecker_product_list : Kronecker product of a list of operators.
        Graph.pauli_kronecker_product : Embed Pauli matrices into a larger Hilbert space.

        Notes
        -----
        This function computes :math:`A_1 \otimes A_2 \otimes ...` where :math:`A_i`
        is either the operator passed for position :math:`i` (if given) or an identity matrix.
        The dimension of :math:`A_i` corresponds to the :math:`i`-th entry in the dimensions list.

        Examples
        --------
        Embed a single operator.

        >>> sigma_z = graph.pauli_matrix("Z")
        >>> graph.embed_operators((sigma_z, 0), [2, 3], name="Z0")
        <Tensor: name="Z0", operation_name="embed_operators", shape=(6, 6)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="Z0")
        >>> result["output"]["Z0"]["value"]
        array([[ 1.+0.j,  0.+0.j,  0.+0.j,  0.+0.j,  0.+0.j,  0.+0.j],
               [ 0.+0.j,  1.+0.j,  0.+0.j,  0.+0.j,  0.+0.j,  0.+0.j],
               [ 0.+0.j,  0.+0.j,  1.+0.j,  0.+0.j,  0.+0.j,  0.+0.j],
               [ 0.+0.j,  0.+0.j,  0.+0.j, -1.+0.j,  0.+0.j,  0.+0.j],
               [ 0.+0.j,  0.+0.j,  0.+0.j,  0.+0.j, -1.+0.j,  0.+0.j],
               [ 0.+0.j,  0.+0.j,  0.+0.j,  0.+0.j,  0.+0.j, -1.+0.j]])

        Embed multiple operators.

        >>> number_op = graph.number_operator(2)
        >>> sigma_x = graph.pauli_matrix("X")
        >>> graph.embed_operators([(number_op, 0), (sigma_x, 2)], [2, 2, 2], name="N0X2")
        <Tensor: name="N0X2", operation_name="embed_operators", shape=(8, 8)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="N0X2")
        >>> result["output"]["N0X2"]["value"]
        array([[0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j]])
        """

        _operators = [operators] if isinstance(operators, tuple) else operators
        for index, (operator, position) in enumerate(_operators):
            Checker.VALUE(
                position < len(dimensions),
                "The position must be less than the length of the dimensions list.",
                {
                    f"operators[{index}][1]": position,
                    "len(dimensions)": len(dimensions),
                },
            )
            Checker.VALUE(
                operator.shape[-2:] == (dimensions[position], dimensions[position]),
                "The last two dimensions of the operator must match the value of the"
                " dimensions list at the given position.",
                {
                    f"shape of {index}-th operator": operator.shape,
                    "position for this operator": position,
                    f"dimensions[{position}]": dimensions[position],
                },
            )

        batch_shape = get_broadcasted_shape(
            *[i[0].shape[:-2] for i in _operators],
            message="The batch shapes of all the operators must be broadcastable.",
        )

        full_dimension = np.prod(dimensions, dtype=int)
        full_shape = batch_shape + (full_dimension, full_dimension)

        operation = create_operation(self.embed_operators, locals())
        return Tensor(operation, shape=full_shape)

    @validated
    def kronecker_product_list(
        self,
        operators: Annotated[
            List[Annotated[Union[np.ndarray, Tensor], pipe(shapeable, after=ShapeT.MATRIX())]],
            Field(min_length=1),
        ],
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        r"""
        Calculate the Kronecker product between a list of operators.

        Parameters
        ----------
        operators : list[Tensor or np.ndarray]
            The list of operators. It must contain at least one operator.
            The operators must be at least 2D and can contain leading batch dimensions.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The Kronecker product of the list of operators. The Hilbert space of the output is equal
            to the tensor product of the Hilbert spaces of all the given operators.

        See Also
        --------
        Graph.embed_operators : Embed operators into a larger Hilbert space.
        Graph.kron : Kronecker product between two objects.
        Graph.pauli_kronecker_product : Embed Pauli matrices into a larger Hilbert space.

        Examples
        --------
        >>> number_op = graph.number_operator(3)
        >>> sigma_x = graph.pauli_matrix("X")
        >>> graph.kronecker_product_list([number_op, sigma_x], name="N0X1")
        <Tensor: name="N0X1", operation_name="kronecker_product_list", shape=(6, 6)>
        >>> result = bo.execute_graph(graph=graph, output_node_names="N0X1")
        >>> result["output"]["N0X1"]["value"]
        array([[0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 2.+0.j],
               [0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 2.+0.j, 0.+0.j]])
        """
        batch_shape = get_broadcasted_shape(
            *[op.shape[:-2] for op in operators],
            message="The batch shapes of the operators must be broadcastable.",
        )
        value_shape = (
            np.prod([op.shape[-2] for op in operators], dtype=int),
            np.prod([op.shape[-1] for op in operators], dtype=int),
        )

        shape = batch_shape + value_shape
        operation = create_operation(self.kronecker_product_list, locals())
        return Tensor(operation, shape=shape)

    @validated
    def min(
        self,
        x: Annotated[Union[np.ndarray, Tensor], pipe(shapeable)],
        axis: Optional[Annotated[Union[int, List[int]], pipe(sequence_like(ScalarT.INT()))]] = None,
        keepdims: bool = False,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Find the minimum value in a tensor along one or multiple of its axes.

        Parameters
        ----------
        x : np.ndarray or Tensor
            The tensor of which you want to find the minimum value.
        axis : int or list[int] or None, optional
            The dimension or dimensions along which you want to search the tensor.
            Defaults to None, in which case this node returns the minimum value
            along all axes of the tensor.
        keepdims : bool, optional
            Whether or not to retain summed axes in the output tensor. If True, each dimension
            in `axis` has size 1 in the result; otherwise, the dimensions in `axis` are
            removed from the result. Defaults to False.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The minimum tensor obtained by searching the input tensor along the specified axes.

        See Also
        --------
        Graph.max : Find the maximum value in a tensor along one or multiple of its axes.
        """
        axes = check_operation_axis(axis, x.shape, "x")
        shape = get_keepdims_operation_shape(x.shape, axes, keepdims)

        operation = create_operation(self.min, locals())
        return Tensor(operation, shape=shape)

    @validated
    def max(
        self,
        x: Annotated[Union[np.ndarray, Tensor], pipe(shapeable)],
        axis: Optional[Annotated[Union[int, List[int]], pipe(sequence_like(ScalarT.INT()))]] = None,
        keepdims: bool = False,
        *,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Find the maximum value in a tensor along one or multiple of its axes.

        Parameters
        ----------
        x : np.ndarray or Tensor
            The tensor of which you want to find the maximum value.
        axis : int or list[int] or None, optional
            The dimension or dimensions along which you want to search the tensor.
            Defaults to None, in which case this node returns the maximum value
            along all axes of the tensor.
        keepdims : bool, optional
            Whether or not to retain summed axes in the output tensor. If True, each dimension
            in `axis` has size 1 in the result; otherwise, the dimensions in `axis` are
            removed from the result. Defaults to False.
        name : str or None, optional
            The name of the node.

        Returns
        -------
        Tensor
            The maximum tensor obtained by searching the input tensor along the specified axes.

        See Also
        --------
        Graph.min : Find the minimum value in a tensor along one or multiple of its axes.
        """
        axes = check_operation_axis(axis, x.shape, "x")
        shape = get_keepdims_operation_shape(x.shape, axes, keepdims)

        operation = create_operation(self.max, locals())
        return Tensor(operation, shape=shape)
